import inspect
from functools import cached_property
from typing import Any, Callable, Optional, Sequence, Union

from pydantic.fields import AliasChoices, AliasPath, FieldInfo

from esmerald.security.scopes import Scopes
from esmerald.typing import Undefined
from esmerald.utils.constants import IS_DEPENDENCY, SKIP_VALIDATION
from esmerald.utils.enums import EncodingType, ParamType
from esmerald.utils.helpers import is_class_and_subclass, make_callable

_PyUndefined: Any = Undefined


class Param(FieldInfo):
    in_: ParamType

    def __init__(
        self,
        default: Any = _PyUndefined,
        *,
        allow_none: Optional[bool] = True,
        default_factory: Optional[Callable[..., Any]] = _PyUndefined,
        annotation: Optional[Any] = None,
        alias: Optional[str] = None,
        alias_priority: Optional[int] = _PyUndefined,
        value_type: Any = _PyUndefined,
        header: Optional[str] = None,
        cookie: Optional[str] = None,
        query: Optional[str] = None,
        path: Optional[str] = None,
        content_encoding: Optional[str] = None,
        required: bool = True,
        title: Optional[str] = None,
        description: Optional[str] = None,
        gt: Optional[float] = None,
        ge: Optional[float] = None,
        lt: Optional[float] = None,
        le: Optional[float] = None,
        multiple_of: Optional[float] = _PyUndefined,
        allow_inf_nan: Optional[bool] = _PyUndefined,
        min_length: Optional[int] = None,
        max_length: Optional[int] = None,
        max_digits: Optional[int] = _PyUndefined,
        strict: Optional[bool] = _PyUndefined,
        pattern: Optional[str] = None,
        examples: Optional[list[Any]] = None,
        deprecated: Optional[bool] = None,
        include_in_schema: bool = True,
        validation_alias: Optional[Union[str, AliasPath, AliasChoices]] = None,
        discriminator: Optional[str] = None,
        frozen: Optional[bool] = None,
        validate_default: bool = True,
        init_var: bool = True,
        kw_only: bool = True,
        json_schema_extra: Optional[dict[str, Any]] = None,
    ) -> None:
        self.deprecated = deprecated
        self.examples = examples
        self.include_in_schema = include_in_schema
        self.allow_none = allow_none

        extra: dict[str, Any] = {}
        extra.update(header=header)
        extra.update(cookie=cookie)
        extra.update(query=query)
        extra.update(path=path)
        extra.update(required=required)
        extra.update(content_encoding=content_encoding)
        extra.update(value_type=value_type)
        extra.update(examples=self.examples)
        extra.update(deprecated=self.deprecated)
        extra.update(include_in_schema=self.include_in_schema)
        extra.update(allow_none=self.allow_none)

        if json_schema_extra is not None:
            extra.update(json_schema_extra)

        super().__init__(
            annotation=annotation,
            default=default,
            default_factory=default_factory,
            alias=alias,
            alias_priority=alias_priority,
            title=title,
            description=description,
            gt=gt,
            ge=ge,
            lt=lt,
            le=le,
            multiple_of=multiple_of,
            min_length=min_length,
            max_length=max_length,
            pattern=pattern,
            examples=examples,
            allow_inf_nan=allow_inf_nan,
            json_schema_extra=extra,
            validate_default=validate_default,
            validation_alias=validation_alias,
            discriminator=discriminator,
            max_digits=max_digits,
            strict=strict,
            frozen=frozen,
            init_var=init_var,
            kw_only=kw_only,
        )

    def __repr__(self) -> str:
        return f"{self.__class__.__name__}(annotation={self.annotation}, default={self.default})"


class Header(Param):
    in_ = ParamType.HEADER

    def __init__(
        self,
        *,
        default: Any = _PyUndefined,
        allow_none: Optional[bool] = True,
        default_factory: Optional[Callable[..., Any]] = _PyUndefined,
        annotation: Optional[Any] = None,
        alias: Optional[str] = None,
        alias_priority: Optional[int] = _PyUndefined,
        value: Optional[str] = None,
        value_type: Any = _PyUndefined,
        content_encoding: Optional[str] = None,
        required: bool = True,
        title: Optional[str] = None,
        description: Optional[str] = None,
        gt: Optional[float] = None,
        ge: Optional[float] = None,
        lt: Optional[float] = None,
        le: Optional[float] = None,
        multiple_of: Optional[float] = _PyUndefined,
        allow_inf_nan: Optional[bool] = _PyUndefined,
        min_length: Optional[int] = None,
        max_length: Optional[int] = None,
        max_digits: Optional[int] = _PyUndefined,
        strict: Optional[bool] = _PyUndefined,
        pattern: Optional[str] = None,
        examples: Optional[list[Any]] = None,
        deprecated: Optional[bool] = None,
        include_in_schema: bool = True,
        validation_alias: Optional[Union[str, AliasPath, AliasChoices]] = None,
        discriminator: Optional[str] = None,
        frozen: Optional[bool] = None,
        validate_default: bool = True,
        init_var: bool = True,
        kw_only: bool = True,
        json_schema_extra: Optional[dict[str, Any]] = None,
    ) -> None:
        super().__init__(
            default=default,
            allow_none=allow_none,
            default_factory=default_factory,
            annotation=annotation,
            header=value,
            alias=alias,
            alias_priority=alias_priority,
            title=title,
            description=description,
            gt=gt,
            ge=ge,
            lt=lt,
            le=le,
            multiple_of=multiple_of,
            allow_inf_nan=allow_inf_nan,
            min_length=min_length,
            max_length=max_length,
            max_digits=max_digits,
            strict=strict,
            pattern=pattern,
            required=required,
            content_encoding=content_encoding,
            value_type=value_type,
            examples=examples,
            deprecated=deprecated,
            include_in_schema=include_in_schema,
            validate_default=validate_default,
            validation_alias=validation_alias,
            discriminator=discriminator,
            frozen=frozen,
            init_var=init_var,
            kw_only=kw_only,
            json_schema_extra=json_schema_extra,
        )


class Cookie(Param):
    in_ = ParamType.COOKIE

    def __init__(
        self,
        default: Any = _PyUndefined,
        *,
        allow_none: Optional[bool] = True,
        default_factory: Optional[Callable[..., Any]] = _PyUndefined,
        annotation: Optional[Any] = None,
        alias: Optional[str] = None,
        alias_priority: Optional[int] = _PyUndefined,
        value_type: Any = _PyUndefined,
        value: Optional[str] = None,
        content_encoding: Optional[str] = None,
        required: bool = True,
        title: Optional[str] = None,
        description: Optional[str] = None,
        gt: Optional[float] = None,
        ge: Optional[float] = None,
        lt: Optional[float] = None,
        le: Optional[float] = None,
        multiple_of: Optional[float] = _PyUndefined,
        allow_inf_nan: Optional[bool] = _PyUndefined,
        min_length: Optional[int] = None,
        max_length: Optional[int] = None,
        max_digits: Optional[int] = _PyUndefined,
        strict: Optional[bool] = _PyUndefined,
        pattern: Optional[str] = None,
        examples: Optional[list[Any]] = None,
        deprecated: Optional[bool] = None,
        include_in_schema: bool = True,
        validation_alias: Optional[Union[str, AliasPath, AliasChoices]] = None,
        discriminator: Optional[str] = None,
        frozen: Optional[bool] = None,
        validate_default: bool = True,
        init_var: bool = True,
        kw_only: bool = True,
        json_schema_extra: Optional[dict[str, Any]] = None,
    ) -> None:
        super().__init__(
            default=default,
            allow_none=allow_none,
            default_factory=default_factory,
            annotation=annotation,
            cookie=value,
            alias=alias,
            alias_priority=alias_priority,
            title=title,
            description=description,
            gt=gt,
            ge=ge,
            lt=lt,
            le=le,
            multiple_of=multiple_of,
            allow_inf_nan=allow_inf_nan,
            min_length=min_length,
            max_length=max_length,
            max_digits=max_digits,
            strict=strict,
            pattern=pattern,
            required=required,
            content_encoding=content_encoding,
            value_type=value_type,
            examples=examples,
            deprecated=deprecated,
            include_in_schema=include_in_schema,
            validate_default=validate_default,
            validation_alias=validation_alias,
            discriminator=discriminator,
            frozen=frozen,
            init_var=init_var,
            kw_only=kw_only,
            json_schema_extra=json_schema_extra,
        )


class Query(Param):
    in_ = ParamType.QUERY

    def __init__(
        self,
        default: Any = _PyUndefined,
        *,
        allow_none: Optional[bool] = True,
        default_factory: Optional[Callable[..., Any]] = _PyUndefined,
        annotation: Optional[Any] = None,
        alias: Optional[str] = None,
        alias_priority: Optional[int] = _PyUndefined,
        value_type: Any = _PyUndefined,
        value: Optional[str] = None,
        content_encoding: Optional[str] = None,
        required: bool = True,
        title: Optional[str] = None,
        description: Optional[str] = None,
        gt: Optional[float] = None,
        ge: Optional[float] = None,
        lt: Optional[float] = None,
        le: Optional[float] = None,
        multiple_of: Optional[float] = _PyUndefined,
        allow_inf_nan: Optional[bool] = _PyUndefined,
        min_length: Optional[int] = None,
        max_length: Optional[int] = None,
        max_digits: Optional[int] = _PyUndefined,
        strict: Optional[bool] = _PyUndefined,
        pattern: Optional[str] = None,
        examples: Optional[list[Any]] = None,
        deprecated: Optional[bool] = None,
        include_in_schema: bool = True,
        validation_alias: Optional[Union[str, AliasPath, AliasChoices]] = None,
        discriminator: Optional[str] = None,
        frozen: Optional[bool] = None,
        validate_default: bool = True,
        init_var: bool = True,
        kw_only: bool = True,
        json_schema_extra: Optional[dict[str, Any]] = None,
    ) -> None:
        super().__init__(
            default=default,
            allow_none=allow_none,
            default_factory=default_factory,
            annotation=annotation,
            query=value,
            alias=alias,
            alias_priority=alias_priority,
            title=title,
            description=description,
            gt=gt,
            ge=ge,
            lt=lt,
            le=le,
            multiple_of=multiple_of,
            allow_inf_nan=allow_inf_nan,
            min_length=min_length,
            max_length=max_length,
            max_digits=max_digits,
            strict=strict,
            pattern=pattern,
            required=required,
            content_encoding=content_encoding,
            value_type=value_type,
            examples=examples,
            deprecated=deprecated,
            include_in_schema=include_in_schema,
            validate_default=validate_default,
            validation_alias=validation_alias,
            discriminator=discriminator,
            frozen=frozen,
            init_var=init_var,
            kw_only=kw_only,
            json_schema_extra=json_schema_extra,
        )


class Path(Param):
    in_ = ParamType.PATH

    def __init__(
        self,
        default: Any = _PyUndefined,
        *,
        allow_none: Optional[bool] = True,
        default_factory: Optional[Callable[..., Any]] = _PyUndefined,
        annotation: Optional[Any] = None,
        value_type: Any = _PyUndefined,
        content_encoding: Optional[str] = None,
        required: bool = True,
        title: Optional[str] = None,
        description: Optional[str] = None,
        gt: Optional[float] = None,
        ge: Optional[float] = None,
        lt: Optional[float] = None,
        le: Optional[float] = None,
        min_length: Optional[int] = None,
        max_length: Optional[int] = None,
        max_digits: Optional[int] = _PyUndefined,
        strict: Optional[bool] = _PyUndefined,
        pattern: Optional[str] = None,
        examples: Optional[list[Any]] = None,
        deprecated: Optional[bool] = None,
        include_in_schema: bool = True,
        validation_alias: Optional[Union[str, AliasPath, AliasChoices]] = None,
        discriminator: Optional[str] = None,
        frozen: Optional[bool] = None,
        validate_default: bool = True,
        init_var: bool = True,
        kw_only: bool = True,
        json_schema_extra: Optional[dict[str, Any]] = None,
    ) -> None:
        super().__init__(
            default=default,
            allow_none=allow_none,
            default_factory=default_factory,
            annotation=annotation,
            title=title,
            description=description,
            gt=gt,
            ge=ge,
            lt=lt,
            le=le,
            min_length=min_length,
            max_length=max_length,
            max_digits=max_digits,
            strict=strict,
            pattern=pattern,
            required=required,
            content_encoding=content_encoding,
            value_type=value_type,
            examples=examples,
            deprecated=deprecated,
            include_in_schema=include_in_schema,
            validate_default=validate_default,
            validation_alias=validation_alias,
            discriminator=discriminator,
            frozen=frozen,
            init_var=init_var,
            kw_only=kw_only,
            json_schema_extra=json_schema_extra,
        )


class Body(FieldInfo):
    def __init__(
        self,
        default: Any = _PyUndefined,
        *,
        allow_none: Optional[bool] = True,
        default_factory: Optional[Callable[..., Any]] = _PyUndefined,
        annotation: Optional[Any] = None,
        media_type: Union[str, EncodingType] = EncodingType.JSON,
        content_encoding: Optional[str] = None,
        title: Optional[str] = None,
        alias: Optional[str] = None,
        alias_priority: Optional[int] = _PyUndefined,
        description: Optional[str] = None,
        embed: Optional[bool] = None,
        gt: Optional[float] = None,
        ge: Optional[float] = None,
        lt: Optional[float] = None,
        le: Optional[float] = None,
        multiple_of: Optional[float] = _PyUndefined,
        allow_inf_nan: Optional[bool] = _PyUndefined,
        max_digits: Optional[int] = _PyUndefined,
        strict: Optional[bool] = _PyUndefined,
        min_length: Optional[int] = None,
        max_length: Optional[int] = None,
        pattern: Optional[str] = None,
        examples: Optional[list[Any]] = None,
        validation_alias: Optional[Union[str, AliasPath, AliasChoices]] = None,
        discriminator: Optional[str] = None,
        frozen: Optional[bool] = None,
        validate_default: bool = True,
        init_var: bool = True,
        kw_only: bool = True,
        include_in_schema: bool = True,
        json_schema_extra: Optional[dict[str, Any]] = None,
    ) -> None:
        extra: dict[str, Any] = {}
        self.media_type = media_type
        self.content_encoding = content_encoding
        self.examples = examples
        self.allow_none = allow_none
        self.include_in_schema = include_in_schema

        extra.update(media_type=self.media_type)
        extra.update(content_encoding=self.content_encoding)
        extra.update(embed=embed)
        extra.update(allow_none=allow_none)

        json_schema_extra = extra if json_schema_extra is None else json_schema_extra.update(extra)

        super().__init__(
            default=default,
            default_factory=default_factory,
            annotation=annotation,
            title=title,
            alias=alias,
            alias_priority=alias_priority,
            description=description,
            gt=gt,
            ge=ge,
            lt=lt,
            le=le,
            multiple_of=multiple_of,
            max_digits=max_digits,
            allow_inf_nan=allow_inf_nan,
            min_length=min_length,
            max_length=max_length,
            pattern=pattern,
            json_schema_extra=json_schema_extra,
            validate_default=validate_default,
            validation_alias=validation_alias,
            discriminator=discriminator,
            frozen=frozen,
            init_var=init_var,
            strict=strict,
            kw_only=kw_only,
        )

    def __repr__(self) -> str:
        return f"{self.__class__.__name__}(annotation={self.annotation}, default={self.default})"


class Form(Body):
    def __init__(
        self,
        default: Any = _PyUndefined,
        *,
        annotation: Optional[Any] = None,
        default_factory: Optional[Callable[..., Any]] = _PyUndefined,
        allow_none: Optional[bool] = True,
        media_type: Union[str, EncodingType] = EncodingType.URL_ENCODED,
        content_encoding: Optional[str] = None,
        alias: Optional[str] = None,
        alias_priority: Optional[int] = _PyUndefined,
        title: Optional[str] = None,
        description: Optional[str] = None,
        gt: Optional[float] = None,
        ge: Optional[float] = None,
        lt: Optional[float] = None,
        le: Optional[float] = None,
        min_length: Optional[int] = None,
        max_length: Optional[int] = None,
        pattern: Optional[str] = None,
        examples: Optional[list[Any]] = None,
        validation_alias: Optional[Union[str, AliasPath, AliasChoices]] = None,
        discriminator: Optional[str] = None,
        max_digits: Optional[int] = _PyUndefined,
        strict: Optional[bool] = _PyUndefined,
        frozen: Optional[bool] = None,
        validate_default: bool = True,
        init_var: bool = True,
        kw_only: bool = True,
        include_in_schema: bool = True,
        json_schema_extra: Optional[dict[str, Any]] = None,
    ) -> None:
        super().__init__(
            default=default,
            annotation=annotation,
            allow_none=allow_none,
            default_factory=default_factory,
            embed=True,
            media_type=media_type,
            content_encoding=content_encoding,
            alias=alias,
            alias_priority=alias_priority,
            title=title,
            description=description,
            gt=gt,
            ge=ge,
            lt=lt,
            le=le,
            min_length=min_length,
            max_length=max_length,
            pattern=pattern,
            examples=examples,
            validate_default=validate_default,
            validation_alias=validation_alias,
            discriminator=discriminator,
            frozen=frozen,
            init_var=init_var,
            kw_only=kw_only,
            include_in_schema=include_in_schema,
            max_digits=max_digits,
            strict=strict,
            json_schema_extra=json_schema_extra,
        )


class File(Form):
    def __init__(
        self,
        default: Any = _PyUndefined,
        *,
        annotation: Optional[Any] = None,
        allow_none: Optional[bool] = True,
        default_factory: Optional[Callable[..., Any]] = _PyUndefined,
        media_type: Union[str, EncodingType] = EncodingType.MULTI_PART,
        content_encoding: Optional[str] = None,
        alias: Optional[str] = None,
        alias_priority: Optional[int] = _PyUndefined,
        title: Optional[str] = None,
        description: Optional[str] = None,
        gt: Optional[float] = None,
        ge: Optional[float] = None,
        lt: Optional[float] = None,
        le: Optional[float] = None,
        min_length: Optional[int] = None,
        max_length: Optional[int] = None,
        pattern: Optional[str] = None,
        examples: Optional[list[Any]] = None,
        validation_alias: Optional[Union[str, AliasPath, AliasChoices]] = None,
        discriminator: Optional[str] = None,
        max_digits: Optional[int] = _PyUndefined,
        strict: Optional[bool] = _PyUndefined,
        frozen: Optional[bool] = None,
        validate_default: bool = True,
        init_var: bool = True,
        kw_only: bool = True,
        include_in_schema: bool = True,
    ) -> None:
        super().__init__(
            default=default,
            annotation=annotation,
            allow_none=allow_none,
            default_factory=default_factory,
            media_type=media_type,
            content_encoding=content_encoding,
            alias=alias,
            alias_priority=alias_priority,
            title=title,
            description=description,
            gt=gt,
            ge=ge,
            lt=lt,
            le=le,
            min_length=min_length,
            max_length=max_length,
            pattern=pattern,
            examples=examples,
            validate_default=validate_default,
            validation_alias=validation_alias,
            discriminator=discriminator,
            max_digits=max_digits,
            strict=strict,
            frozen=frozen,
            init_var=init_var,
            kw_only=kw_only,
            include_in_schema=include_in_schema,
        )


class Injects(FieldInfo):
    """
    Creates a FieldInfo class with extra parameters.
    This is used for dependencies and to inject them.

    **Example**

    ```python
    @get(dependencies={"value": Inject(lambda: 13)})
    def myview(value: Injects()):
        return {"value": value}
    ```
    """

    def __init__(
        self,
        default: Any = Undefined,
        skip_validation: bool = False,
        allow_none: bool = True,
    ) -> None:
        self.allow_none = allow_none
        self.extra: dict[str, Any] = {
            IS_DEPENDENCY: True,
            SKIP_VALIDATION: skip_validation,
            "allow_none": self.allow_none,
        }
        super().__init__(default=default, json_schema_extra=self.extra)


class BaseRequires:
    """
    A class that represents a requirement with an optional dependency and caching behavior.

    This object serves as a base class for other classes that require dependencies.

    Attributes:
        dependency (Optional[Callable[..., Any]]): An optional callable that represents the dependency.
        use_cache (bool): A flag indicating whether to use caching for the dependency. Defaults to True.

    Methods:
        __repr__(): Returns a string representation of the Requires instance.
    """

    def __init__(self, dependency: Optional[Callable[..., Any]] = None, *, use_cache: bool = True):
        """
        Initializes a Requires instance.

        Args:
            dependency (Optional[Callable[..., Any]]): An optional callable that represents the dependency.
            use_cache (bool): A flag indicating whether to use caching for the dependency. Defaults to True.
        """

        """
        Returns a string representation of the Requires instance.

        Returns:
            str: A string representation of the Requires instance.
        """
        self.dependency = dependency
        self.use_cache = use_cache

        if not callable(dependency):
            dependency = make_callable(dependency)
        self.signature_model = inspect.signature(dependency) if dependency else None

    def __repr__(self) -> str:
        attr = getattr(self.dependency, "__name__", type(self.dependency).__name__)
        cache = "" if self.use_cache else ", use_cache=False"
        return f"{self.__class__.__name__}({attr}{cache})"


class Requires(BaseRequires): ...


class Security(BaseRequires):
    """
    A class used to represent security requirements for a particular operation.

    Attributes:
    ----------
    dependency : Optional[Callable[..., Any]]
        A callable that represents the dependency required for security.
    scopes : Optional[Sequence[str]]
        A sequence of scopes required for the security. Defaults to an empty list.
    use_cache : bool
        A flag indicating whether to use cache. Defaults to True.

    Methods:
    -------
    __init__(self, dependency: Optional[Callable[..., Any]] = None, *, scopes: Optional[Sequence[str]] = None, use_cache: bool = True)
        Initializes the Security class with the given dependency, scopes, and use_cache flag.
    """

    def __init__(
        self,
        dependency: Optional[Callable[..., Any]] = None,
        *,
        scopes: Optional[Sequence[str]] = None,
        use_cache: bool = True,
    ):
        super().__init__(dependency=dependency, use_cache=use_cache)
        self.scopes = scopes or []

    @cached_property
    def is_security_scope_dependency(self) -> bool:
        parameters: dict[str, inspect.Parameter] = dict(self.signature_model.parameters.items())
        return any(
            is_class_and_subclass(param.annotation, Scopes) for param in parameters.values()
        )



from __future__ import annotations

from lilya.types import ASGIApp

from esmerald.contrib.grpc.gateway import GrpcGateway


def register_grpc_http_routes(
    app: ASGIApp, grpc_gateways: list[GrpcGateway], base_path: str | None = None
) -> None:
    """
    Registers HTTP routes corresponding to gRPC service methods to the Esmerald app.

    This function takes a gRPC gateway (GrpcGateway) instance that maps gRPC service
    methods to HTTP routes and registers these HTTP routes to the provided Esmerald app.

    The HTTP routes are dynamically created based on the service methods in the
    GrpcGateway instance, and the appropriate handler is attached for each HTTP method
    (GET, POST, PUT, DELETE, PATCH). The handler is a wrapper that invokes the
    corresponding gRPC method and returns the result as a JSON response.

    Args:
        app (ASGIApp): The Esmerald app instance where the HTTP routes will be registered.
        grpc_gateways (list[GrpcGateway]): The GrpcGateway list of instances that contains the gRPC services
                                    and their method mappings to HTTP routes.
        base_path (str, optional): The base path for the HTTP routes. Defaults to None.


    Yields:
        None: This function does not return any values, but it registers routes in-place.
    """
    if not base_path:
        base_path = "/"

    for grpc_gateway in grpc_gateways:
        for handler in grpc_gateway._register_http_endpoints():
            app.add_route(base_path, handler)



from abc import ABC

from lilya.logging import LoggingConfig as LilyaLoggingConfig


class LoggingConfig(LilyaLoggingConfig, ABC):
    """
    An instance of [LoggingConfig](https://esmerald.dev/configurations/logging/).

    !!! Tip
        You can create your own `LoggingMiddleware` version and pass your own
        configurations. You don't need to use the built-in version although it
        is recommended to do it so.

    **Example**

    ```python
    from esmerald import Esmerald
    from esmerald.core.config import LoggingConfig

    logging_config = LoggingConfig()

    app = Esmerald(logging_config=logging_config)
    ```
    """

    ...



from decimal import Decimal
from typing import Any, Type, TypeVar, Union, _GenericAlias, cast, get_args

from pydantic.fields import FieldInfo
from pydantic.json_schema import SkipJsonSchema

T = TypeVar("T", int, float, Decimal)


def is_field_optional(field: "FieldInfo") -> bool:
    """
    Returns bool True or False for the optional model field.
    """
    allow_none = getattr(field, "allow_none", True)
    return not field.is_required() and not is_any_type(field=field) and allow_none


def is_any_type(field: "FieldInfo") -> bool:
    """
    Checks if the field is of type Any.
    """
    name = cast("Any", getattr(field.annotation, "_name", None))
    return (name is not None and "Any" in name) or field.annotation is Any


def should_skip_json_schema(field_info: Union[FieldInfo, Any]) -> FieldInfo:
    """
    Checks if the schema generation for the parameters should be skipped.
    This is applied for complex fields in query params.

    Example:
        1. Union[dict[str, str], None]
        2. Optional[dict[str, str]]
        3. Union[list[str], None]
        4. Optional[list[str]]
    """
    union_args = get_args(field_info.annotation)

    if not any(
        isinstance(annotation, _GenericAlias) for annotation in union_args
    ) or not isinstance(field_info.annotation, _GenericAlias):
        return field_info

    arguments: list[Type[Any]] = []

    for argument in union_args:
        if argument is not type(None):
            arguments.append(argument)
        else:
            arguments.append(SkipJsonSchema[None])

    arguments = tuple(arguments)  # type: ignore
    field_info.annotation = Union[arguments]
    return field_info


def extract_arguments(
    param: Union[Any, None] = None, argument_list: Union[list[Any], None] = None
) -> list[Type[type]]:
    """
    Recursively extracts unique types from a parameter's type annotation.

    Args:
        param (Union[Parameter, None], optional): The parameter with type annotation to extract from.
        argument_list (Union[list[Any], None], optional): The list of arguments extracted so far.

    Returns:
        list[Type[type]]: A list of unique types extracted from the parameter's type annotation.
    """
    arguments: list[Any] = list(argument_list) if argument_list is not None else []
    args = get_args(param)

    for arg in args:
        if isinstance(arg, _GenericAlias):
            arguments.extend(extract_arguments(param=arg, argument_list=arguments))
        else:
            if arg not in arguments:
                arguments.append(arg)
    return arguments



import inspect
from typing import Any, Union

from lilya.compat import run_sync
from lilya.context import request_context

from esmerald import params
from esmerald.security.scopes import Scopes
from esmerald.utils.helpers import is_class_and_subclass


def is_requires_scheme(param: Any) -> bool:
    """
    Checks if the object is a security scheme.
    """
    return is_class_and_subclass(param, params.Requires)


def is_security_scheme(param: Any) -> bool:
    """
    Checks if the object is a security scheme.
    """
    if not param:
        return False
    return isinstance(param, params.Security)


def is_base_requires(param: Any) -> bool:
    """
    Checks if the object is a base requires object.
    """
    return is_class_and_subclass(param, params.BaseRequires)


def is_security_scope(param: Any) -> bool:
    """
    Checks if the object is a security scope object.
    """
    if not param:
        return False
    return is_class_and_subclass(param, Scopes)


def is_inject(param: Any) -> bool:
    """
    Checks if the object is an Inject.
    """
    from esmerald.injector import Inject

    return isinstance(param, Inject)


def is_requires(param: Any) -> bool:
    """
    Checks if the object is an Inject.
    """
    if not param:
        return False
    return isinstance(param, params.Requires)


async def async_resolve_dependencies(func: Any, overrides: Union[dict[str, Any]] = None) -> Any:
    """
    Resolves dependencies for an asynchronous function by inspecting its signature and
    recursively resolving any dependencies specified using the `params.Requires` class.
    Args:
        func (Any): The target function whose dependencies need to be resolved.
        overrides (Union[dict[str, Any]], optional): A dictionary of overrides for dependencies.
            This can be used for testing or customization. Defaults to None.
    Returns:
        Any: The result of the target function with its dependencies resolved.
    Raises:
        TypeError: If the target function or any of its dependencies are not callable.
    """
    if overrides is None:
        overrides = {}

    signature = inspect.signature(func)
    kwargs = {}

    for name, param in signature.parameters.items():
        # If in one of the requirements happens to be Security, we need to resolve it
        # By passing the Request object to the dependency
        if isinstance(param.default, params.Security):
            kwargs[name] = await param.default.dependency(request_context)
        if isinstance(param.default, params.Requires):
            dep_func = param.default.dependency
            dep_func = overrides.get(dep_func, dep_func)  # type: ignore
            if inspect.iscoroutinefunction(dep_func):
                resolved = await async_resolve_dependencies(dep_func, overrides)
            else:
                resolved = (
                    resolve_dependencies(dep_func, overrides) if callable(dep_func) else dep_func
                )
            kwargs[name] = resolved
    if inspect.iscoroutinefunction(func):
        return await func(**kwargs)
    else:
        return func(**kwargs)


def resolve_dependencies(func: Any, overrides: Union[dict[str, Any]] = None) -> Any:
    """
    Resolves the dependencies for a given function.

    Parameters:
        func (Any): The function for which dependencies need to be resolved.
        overrides (Union[dict[str, Any], None], optional): A dictionary of dependency overrides. Defaults to None.
        Raises:
        ValueError: If the provided function is asynchronous.

    Returns:
        Any: The result of running the asynchronous dependency resolution function.
    """
    if overrides is None:
        overrides = {}
    if inspect.iscoroutinefunction(func):
        raise ValueError("Function is async. Use resolve_dependencies_async instead.")
    return run_sync(async_resolve_dependencies(func, overrides))



from __future__ import annotations

from typing import Any

try:
    import grpc
except ImportError:
    raise ImportError(
        "You must have the official `grpcio-tools` installed to use this module."
    ) from None

from google.protobuf.json_format import MessageToDict
from grpc import aio
from lilya._internal._path import clean_path

from esmerald import HTTPException, Request, route
from esmerald.exceptions import ImproperlyMiddlewareConfigured
from esmerald.injector import Inject
from esmerald.params import Injects
from esmerald.responses import JSONResponse
from esmerald.utils.helpers import make_callable

# Allowed HTTP methods for endpoints.
HTTP_ALLOWED_METHODS = ["GET", "POST", "PUT", "PATCH", "DELETE", "OPTIONS", "TRACE"]


class HTTPContext:
    """
    A minimal dummy gRPC context for HTTP calls.
    It implements only the methods required to satisfy the service signature.
    """

    def __init__(self) -> None:
        self._code = grpc.StatusCode.OK
        self._details = ""

    def set_code(self, code: grpc.StatusCode) -> None:
        self._code = code

    def set_details(self, details: str) -> None:
        self._details = details

    @property
    def code(self) -> grpc.StatusCode:
        return self._code

    @property
    def details(self) -> str:
        return self._details


class GrpcGateway:
    """
    GrpcGateway bridges gRPC services (using grpcio) and HTTP endpoints in an Esmerald app.
    It registers the gRPC services with a grpc.aio.Server and (optionally) exposes HTTP endpoints
    that call the service methods directly.

    Attributes:
        services (list[type]): list of gRPC service classes.
        grpc_server (aio.Server): The grpc.aio server instance.
        path (str): The base HTTP path used for generating endpoints.
        server_options (dict): Options such as host and port for the gRPC server.
        http_methods (list[str]): Allowed HTTP methods for the endpoints.
    """

    def __init__(
        self,
        path: str,
        services: list[type],
        expose_http: bool = True,
        http_methods: list[str] | None = None,
        is_secure: bool = False,
        server_credentials: grpc.ServerCredentials | None = None,
        **server_options: Any,
    ) -> None:
        """
        Initializes the GrpcGateway instance.

        Args:
            path (str): Base HTTP path (must start with '/').
            services (list[type]): list of gRPC service classes to expose.
            expose_http (bool): Whether to register HTTP endpoints that map to gRPC methods.
            http_methods (list[str] | None): If provided, limits the HTTP methods allowed.
            **server_options (Any): Additional options (e.g., host and port) for the gRPC server.

        Raises:
            AssertionError: If the provided path does not start with '/'.
        """
        assert path.startswith("/"), "Paths must start with '/'"

        self.path = clean_path(path)
        self.services = services
        # Create a grpc.aio.Server instance.
        self.grpc_server = aio.server()
        self.server_options = server_options
        self.http_methods = (
            [method.upper() for method in http_methods]
            if http_methods is not None
            else HTTP_ALLOWED_METHODS
        )

        # Register each service with the grpc.aio server.
        for service in services:
            instance = service()
            # We expect each service type to have a callable '__add_to_server__'
            # which registers the service with a grpc.aio.Server.
            add_fn = getattr(service, "__add_to_server__", None)
            if add_fn is None or not callable(add_fn):
                raise ValueError(
                    f"Service {service.__name__} must provide a '__add_to_server__' callable."
                )
            add_fn(instance, self.grpc_server)

        if expose_http:
            self._register_http_endpoints()

        # Set default host/port if not provided.
        if "host" not in self.server_options:
            self.server_options["host"] = "127.0.0.1"
        if "port" not in self.server_options:
            self.server_options["port"] = 50051

        self.is_secure = is_secure
        self.server_credentials = server_credentials

        if self.is_secure and not self.server_credentials:
            raise ImproperlyMiddlewareConfigured(
                "When `is_secure` is true, you must provide the `server_credentials`."
            )

    async def startup(self) -> None:
        """
        Starts the grpc.aio server asynchronously.
        """
        host = self.server_options["host"]
        port = self.server_options["port"]

        if not self.is_secure:
            self.grpc_server.add_insecure_port(f"{host}:{port}")
        else:
            self.grpc_server.add_secure_port(
                f"{host}:{port}", server_credentials=self.server_credentials
            )
        await self.grpc_server.start()

    async def shutdown(self) -> None:
        """
        Shuts down the grpc.aio server asynchronously.
        """
        await self.grpc_server.stop(0)

    def _register_http_endpoints(self) -> Any:
        """
        Registers HTTP endpoints for each gRPC service method.

        For each service in `self.services`, this method dynamically generates HTTP routes and associates
        them with the corresponding gRPC service method. It uses Esmerald decorators to define the HTTP
        methods (`GET`, `POST`, etc.), maps them to gRPC methods, and defines a handler that wraps the
        gRPC method with HTTP request handling.

        This method assumes that each service has callable methods that should be exposed as HTTP routes.
        """
        for service in self.services:
            service_instance = service()

            for method_name in dir(service_instance):
                if method_name.startswith("_"):
                    continue  # Skip private methods

                method = getattr(service_instance, method_name)
                if callable(method):
                    route_path = clean_path(
                        f"/{self.path}/{service.__name__.lower()}/{method_name.lower()}"
                    )
                    name = route_path.replace("/", "_").replace(".", "_").replace("__", "_")

                    if name.startswith("_"):
                        name = name[1:]

                    # Define a dependency for the service instance
                    def service_dependency() -> Any:
                        return service_instance  # noqa

                    callable_fn = make_callable(method_name)

                    @route(
                        path=route_path,
                        tags=["gRPC"],
                        methods=self.http_methods,
                        name=name,
                        dependencies={
                            "service": Inject(service_dependency),
                            "service_method": Inject(callable_fn),
                        },
                    )
                    async def http_wrapper(
                        request: Request,
                        service: Any = Injects(),
                        service_method: Any = Injects(),
                    ) -> Any:
                        """
                        HTTP handler that wraps the gRPC method. It receives HTTP requests, converts them into
                        data that can be processed by the gRPC method, and returns the result as an HTTP response.

                        Args:
                            request (Request): The HTTP request object containing input data.
                            service (Any): The gRPC service instance injected from dependencies.
                            service_method (Any): The gRPC method to be invoked, injected from dependencies.

                        Returns:
                            JSONResponse: The response that will be sent back to the client, wrapped as a JSON response.

                        Raises:
                            HTTPException: If an error occurs while invoking the gRPC method, this exception is raised
                            with the appropriate HTTP status code and error message.
                        """
                        method = getattr(service, service_method)

                        try:
                            data = await request.json()  # Expecting a dict.
                            http_context = HTTPContext()  # Create a dummy context.
                            response = await method(
                                data, http_context
                            )  # Call the gRPC method with (data, context).
                            # If response is a protobuf message, convert it to dict.
                            if http_context.code != grpc.StatusCode.OK:
                                raise HTTPException(
                                    status_code=self.__grpc_to_http_status(http_context.code),
                                    detail=http_context.details,
                                )

                            if hasattr(response, "SerializeToString"):
                                response = MessageToDict(
                                    response, preserving_proto_field_name=True
                                )
                            return JSONResponse(response)
                        except grpc.aio.AioRpcError as e:
                            raise HTTPException(
                                status_code=self.__grpc_to_http_status(e.code()),
                                detail=e.details(),
                            ) from e

                    # Yield the HTTP method, route, and handler for use by the framework
                    http_name = f"http_handler_{service_instance.__class__.__name__.lower()}_{method.__name__.lower()}"
                    http_wrapper.fn.__name__ = http_name
                    yield http_wrapper

    def __grpc_to_http_status(self, grpc_status: grpc.StatusCode) -> int:
        """
        Converts gRPC status codes to HTTP status codes.

        This method maps common gRPC status codes to their corresponding HTTP status codes. This mapping is used
        to ensure that gRPC errors are communicated effectively as HTTP responses.

        Args:
            grpc_status (Status): The gRPC status code to be converted to an HTTP status code.

        Returns:
            int: The corresponding HTTP status code. Defaults to 500 (Internal Server Error) if the status code
                 is not mapped.
        """
        mapping = {
            grpc.StatusCode.OK: 200,
            grpc.StatusCode.CANCELLED: 499,
            grpc.StatusCode.UNKNOWN: 500,
            grpc.StatusCode.INVALID_ARGUMENT: 400,
            grpc.StatusCode.DEADLINE_EXCEEDED: 504,
            grpc.StatusCode.NOT_FOUND: 404,
            grpc.StatusCode.ALREADY_EXISTS: 409,
            grpc.StatusCode.PERMISSION_DENIED: 403,
            grpc.StatusCode.UNAUTHENTICATED: 401,
            grpc.StatusCode.RESOURCE_EXHAUSTED: 429,
            grpc.StatusCode.FAILED_PRECONDITION: 400,
            grpc.StatusCode.ABORTED: 409,
            grpc.StatusCode.OUT_OF_RANGE: 400,
            grpc.StatusCode.UNIMPLEMENTED: 501,
            grpc.StatusCode.INTERNAL: 500,
            grpc.StatusCode.UNAVAILABLE: 503,
            grpc.StatusCode.DATA_LOSS: 500,
        }
        return mapping.get(grpc_status, 500)  # Default to 500 if the status code is unknown



from typing import TYPE_CHECKING, Any, Callable, Optional, Union, cast

from lilya.permissions.base import DefinePermission

from esmerald.exceptions import PermissionDenied
from esmerald.utils.helpers import is_async_callable, is_class_and_subclass
from esmerald.utils.sync import AsyncCallable

if TYPE_CHECKING:  # pragma: no cover
    from esmerald.permissions import BasePermission
    from esmerald.requests import Request
    from esmerald.types import APIGateHandler


async def continue_or_raise_permission_exception(
    request: "Request",
    apiview: "APIGateHandler",
    permission: "BasePermission",
) -> None:
    """
    Check if the request has permission to access the API view.
    If not permitted, raise a PermissionDenied exception.
    """
    has_permission: Callable = permission.has_permission

    if not is_async_callable(has_permission):
        if not has_permission(request=request, apiview=apiview):
            permission_denied(
                request,
                message=getattr(permission, "message", None),
            )
    else:
        is_permission = await has_permission(request=request, apiview=apiview)
        if not is_permission:
            permission_denied(
                request,
                message=getattr(permission, "message", None),
            )


def permission_denied(request: "Request", message: Optional[str] = None) -> None:
    """
    If request is not permitted, determine what kind of exception to raise.
    """
    raise PermissionDenied(detail=message, status_code=403)


def is_esmerald_permission(permission: Union["BasePermission", Any]) -> bool:
    """
    Checks if the given permission is an instance or subclass of BasePermission.

    Args:
        permission (Union["BasePermission", Any]): The permission to check.
    Returns:
        bool: True if the permission is an instance or subclass of BasePermission, False otherwise.
    """

    from esmerald.permissions import BasePermission

    return is_class_and_subclass(permission, BasePermission)


def wrap_permission(
    permission: Union[AsyncCallable, DefinePermission, Any],
) -> Union["BasePermission", DefinePermission]:
    """
    Wraps the given permission into a BasePermission instance if it is not already one.
    Or else it will assume its a Lilya permission and wraps it.

    Args:
        permission (Union["BasePermission", Any]): The permission to be wrapped.
    Returns:
        BasePermission: The wrapped permission instance.
    """
    if is_esmerald_permission(permission):
        return cast("BasePermission", AsyncCallable(permission))

    # If its an instance of a DefinePermission, then return it.
    if isinstance(permission, DefinePermission):
        return permission
    return DefinePermission(cast(Any, permission))



from __future__ import annotations

from abc import ABC, abstractmethod
from inspect import isclass
from typing import TYPE_CHECKING, Any, Iterator, Optional, cast

from monkay import load
from typing_extensions import Annotated, Doc

from esmerald.core.protocols.extension import ExtensionProtocol
from esmerald.exceptions import ImproperlyConfigured

if TYPE_CHECKING:  # pragma: no cover
    from esmerald.applications import Esmerald


class Pluggable:
    """
    The `Pluggable` is a wrapper around an Extension to initialize it lazily.

    Read more about the [Pluggables](https://esmerald.dev/pluggables/) and learn how to use them.

    **Example**

    ```python
    from typing import Optional

    from loguru import logger
    from pydantic import BaseModel

    from esmerald import Esmerald, Extension, Pluggable
    from esmerald.types import DictAny


    class PluggableConfig(BaseModel):
        name: str


    class MyExtension(Extension):
        def __init__(
            self,
            app: Optional["Esmerald"] = None,
            config: PluggableConfig = None,
            **kwargs: "DictAny",
        ):
            super().__init__(app, **kwargs)

        def extend(self, config: PluggableConfig) -> None:
            logger.success(f"Successfully passed a config {config.name}")


    my_config = PluggableConfig(name="my extension")

    pluggable = Pluggable(MyExtension, config=my_config)
    # or
    # pluggable = Pluggable("path.to.MyExtension", config=my_config)

    app = Esmerald(routes=[], extensions={"my-extension": pluggable})
    ```
    """

    def __init__(self, cls: type[ExtensionProtocol] | str, **options: Any):
        self.cls_or_string = cls
        self.options = options

    @property
    def cls(self) -> type[ExtensionProtocol]:
        cls_or_string = self.cls_or_string
        if isinstance(cls_or_string, str):
            self.cls_or_string = cls_or_string = load(cls_or_string)
        return cast(type["ExtensionProtocol"], cls_or_string)

    def __iter__(self) -> Iterator:
        iterator = (self.cls, self.options)
        return iter(iterator)

    def __repr__(self) -> str:  # pragma: no cover
        name = self.__class__.__name__
        options = [f"{key}={value!r}" for key, value in self.options.items()]
        args = ", ".join([self.__class__.__name__] + options)
        return f"{name}({args})"


class Extension(ABC, ExtensionProtocol):
    """
    `Extension` object is the one being used to add the logic that will originate
    the `pluggable` in the application.

    The `Extension` **must implement** the `extend` function.

    Read more about the [Extension](https://esmerald.dev/pluggables/#extension) and learn how to use it.

    **Example**

    ```python
    from typing import Optional

    from esmerald import Esmerald, Extension
    from esmerald.types import DictAny


    class MyExtension(Extension):
        def __init__(self, app: Optional["Esmerald"] = None, **kwargs: "DictAny"):
            super().__init__(app)
            self.kwargs = kwargs

        def extend(self, **kwargs: "DictAny") -> None:
            '''
            Function that should always be implemented when extending
            the Extension class or a `NotImplementedError` is raised.
            '''
            # Do something here
    ```
    """

    def __init__(
        self,
        app: Annotated[
            Optional[Esmerald],
            Doc(
                """
                An `Esmerald` application instance or subclasses of Esmerald.
                """
            ),
        ] = None,
        **kwargs: Annotated[
            Any,
            Doc("""Any additional kwargs needed."""),
        ],
    ):
        super().__init__()
        self.app = app

    @abstractmethod
    def extend(self, **kwargs: Any) -> None:
        raise NotImplementedError("Extension must be implemented by the subclasses.")


BaseExtension = Extension


class ExtensionDict(dict[str, Extension]):
    def __init__(self, data: Any = None, *, app: Esmerald):
        super().__init__(data)
        self.app = app
        self.delayed_extend: Optional[dict[str, dict[str, Any]]] = {}
        for k, v in self.items():
            self[k] = v

    def extend(self) -> None:
        while self.delayed_extend:
            key, val = self.delayed_extend.popitem()
            self[key].extend(**val)
        self.delayed_extend = None

    def ensure_extension(self, name: str) -> None:
        """
        For reordering extension initialization

        class MyExtension2(Extension):

            def extend(self, **kwargs: "DictAny") -> None:
                '''
                Function that should always be implemented when extending
                the Extension class or a `NotImplementedError` is raised.
                '''
                self.app.extensions.ensure_extension("foo")
                # Do something here
        ```

        """
        if name not in self:
            raise ValueError(f"Extension does not exist: {name}")
        delayed = self.delayed_extend
        if delayed is None:
            return
        val = delayed.pop(name, None)
        if val is not None:
            self[name].extend(**val)

    def __setitem__(self, name: Any, value: Any) -> None:
        if not isinstance(name, str):
            raise ImproperlyConfigured("Extension names should be in string format.")

        if isinstance(value, str):
            value = Pluggable(value)

        if isinstance(value, Pluggable):
            cls, options = value
            value = cls(app=self.app, **options)
            if self.delayed_extend is None:
                value.extend(**options)
            else:
                self.delayed_extend[name] = options
        elif isclass(value) and issubclass(value, ExtensionProtocol):
            value = value(app=self.app)
            if self.delayed_extend is None:
                value.extend()
            else:
                self.delayed_extend[name] = {}
        elif isinstance(value, ExtensionProtocol):
            if self.delayed_extend is not None:
                raise ImproperlyConfigured(
                    "Cannot pass an initialized extension in extensions parameter."
                )
        else:
            raise ImproperlyConfigured(
                "An extension must subclass from Extension, implement the ExtensionProtocol "
                "as instance or being wrapped in a Pluggable."
            )

        super().__setitem__(name, value)



from typing import Optional

from esmerald.openapi.models import (
    SecurityScheme,
)


class SecurityBase(SecurityScheme):
    scheme_name: Optional[str] = None
    """
    An optional name for the security scheme.
    """
    __auto_error__: bool = False
    """
    A flag to indicate if automatic error handling should be enabled.
    """
    __is_security__: bool = True
    """A flag to indicate that this is a security scheme. """


class HttpSecurityBase(SecurityScheme):
    scheme_name: Optional[str] = None
    """
    An optional name for the security scheme.
    """
    realm: Optional[str] = None
    """
    An optional realm for the security scheme.
    """
    __auto_error__: bool = False
    """
    A flag to indicate if automatic error handling should be enabled.
    """



from collections.abc import Callable
from typing import Any


def directive(func: Callable[..., Any]) -> Callable:
    """
    Marks a function-based Sayer CLI command as a custom Esmerald directive.
    This decorator is used to register user-defined CLI commands that follow the
    function-based directive pattern. It tags the command with a special attribute
    (__is_custom_directive__ = True) so Esmerald's CLI system can treat it differently
    from internal class-based directives.
    The command must already be decorated with `@command` from Sayer before applying this.
    Example usage:
        @directive
        @command(name="create")
        async def create(name: Annotated[str, Option(help="Your name")]):
            ...
    Returns:
        Callable: The original command object, with a marker and registered in the main CLI.
    """
    from esmerald.core.directives.cli import esmerald_cli

    func.__is_custom_directive__ = True
    func.get_help = esmerald_cli.get_help
    esmerald_cli.add_command(func)

    return func



from datetime import datetime
from typing import (
    TYPE_CHECKING,
    Any,
    Awaitable,
    Callable,
    Mapping,
    Type,
    TypeVar,
    Union,
    get_args,
)

from lilya.middleware import DefineMiddleware
from lilya.types import ASGIApp
from typing_extensions import Literal

from esmerald.background import BackgroundTask, BackgroundTasks
from esmerald.exceptions import MissingDependency
from esmerald.routing.gateways import WebSocketGateway
from esmerald.routing.router import Include

try:
    from esmerald.core.config.template import TemplateConfig as TemplateConfig
except MissingDependency:
    TemplateConfig = Any  # type: ignore

if TYPE_CHECKING:
    from esmerald.applications import Application, Esmerald
    from esmerald.conf.global_settings import EsmeraldSettings  # noqa
    from esmerald.core.datastructures import Cookie, ResponseHeader
    from esmerald.injector import Inject
    from esmerald.requests import Request
    from esmerald.responses import Response
    from esmerald.routing.apis.base import View
    from esmerald.routing.gateways import Gateway, WebhookGateway  # noqa
    from esmerald.routing.router import (
        HTTPHandler as HTTPHandler,  # noqa
        Router,
        WebSocketHandler as WebSocketHandler,  # noqa
    )
    from esmerald.websockets import WebSocket  # noqa

    # prevent ciruclar and unneccessary imports
    # by only exposing it when type checking
    # also only expose if available otherwise fallback to Any
    try:
        from asyncz.schedulers.types import SchedulerType as SchedulerType  # noqa
    except ImportError:
        SchedulerType = Any  # type: ignore
else:
    HTTPHandler = Any
    Application = Any
    Message = Any
    Receive = Any
    Scope = Any
    Send = Any
    BaseHTTPMiddleware = Any
    Inject = Any
    Request = Any
    WebSocket = Any
    State = Any
    Response = Any
    ResponseHeader = Any
    Cookie = Any
    Router = Any
    MiddlewareProtocol = Any
    View = Any
    Gateway = Any
    WebhookGateway = Any
    Esmerald = Any
    EsmeraldSettings = Any
    SchedulerType = Any

AsyncAnyCallable = Callable[..., Awaitable[Any]]
HTTPMethod = Literal["GET", "POST", "DELETE", "PATCH", "PUT", "HEAD"]
HTTPMethods = get_args(HTTPMethod)

Middleware = DefineMiddleware

ResponseType = Type[Response]

Dependencies = Union[dict[str, Inject], dict[str, Any], Any]

ExceptionType = TypeVar("ExceptionType", bound=Exception)
ExceptionHandler = Callable[[Request, ExceptionType], Response]
ExceptionHandlerMap = Mapping[Union[int, Type[Exception]], ExceptionHandler]

_ReservedKwargs = Literal[
    "request", "socket", "headers", "query", "cookies", "state", "data", "payload"
]

ReservedKwargs = get_args(_ReservedKwargs)

ResponseHeaders = dict[str, ResponseHeader]
ResponseCookies = list[Cookie]
AsyncAnyCallable = Callable[..., Awaitable[Any]]

DatetimeType = TypeVar("DatetimeType", bound=datetime)

ParentType = Union[View, Router, Gateway, WebSocketGateway, Esmerald, Include, Application]
APIGateHandler = Union[
    Gateway,
    WebSocketGateway,
]

RouteParent = Union["Router", "Include", ASGIApp, "Gateway", "WebSocketGateway", "WebhookGateway"]

BackgroundTaskType = Union[BackgroundTask, BackgroundTasks]
SecurityScheme = dict[str, list[str]]

ConnectionType = Union["Request", "WebSocket"]
DictStr = dict[str, str]
DictAny = dict[str, Any]
SettingsType = Type["EsmeraldSettings"]

LifeSpanHandler = Union[Callable[[], None], Callable[[], Awaitable[None]]]



"""
Functions to use with the Factory dependency injection.
"""

from importlib import import_module
from typing import Any, Callable, Tuple, cast

from monkay import load  # noqa

from esmerald.exceptions import ImproperlyConfigured


def _resolve_imported_attribute(target: str, attribute: str) -> Any:  # pragma: no cover
    """
    Dynamically imports a module and retrieves the requested attribute.

    This function efficiently resolves deeply nested attributes by progressively importing
    each module level while avoiding redundant re-imports.

    Args:
        target (str): The fully qualified module path (e.g., "module.something").
        attribute (str): The attribute name to retrieve from the module.

    Returns:
        Any: The resolved attribute.

    Raises:
        ImportError: If the module or attribute cannot be imported.
    """
    try:
        module = import_module(target)
        return getattr(module, attribute)
    except ModuleNotFoundError:
        # If `target` is not a module, progressively resolve attributes.
        components = target.split(".")
        module_name = components.pop(0)
        module = __import__(module_name)

        for comp in components:
            module_name += f".{comp}"
            try:
                module = getattr(module, comp)
            except AttributeError:
                module = import_module(module_name)

        return getattr(module, attribute)
    except AttributeError as e:
        raise ImportError(f"Module '{target}' has no attribute '{attribute}'.") from e


def _get_provider_callable(provider: str) -> Callable[[], Any]:  # pragma: no cover
    """
    Returns a lazy loader function for a given provider string.

    Args:
        provider (str): The fully qualified import path (e.g., "myapp.daos.MyDAO").

    Returns:
        Callable[[], Any]: A function that, when called, imports and returns the attribute.

    Raises:
        TypeError: If the provider string is not in the correct format.
    """
    try:
        target, attribute = provider.rsplit(".", 1)
    except (ValueError, AttributeError, TypeError):
        raise TypeError(
            f"Invalid provider format: {provider!r}. Expected 'module.attribute'."
        ) from None

    return lambda: _resolve_imported_attribute(target, attribute)


def load_provider(provider: str) -> Tuple[Callable, bool]:  # pragma: no cover
    """
    Dynamically loads a callable provider from a string reference.

    This function allows for dependency injection without explicit imports,
    enabling modular and scalable application architecture.

    Args:
        provider (str): The fully qualified import path (e.g., "myapp.daos.MyDAO").

    Returns:
        Tuple[Callable, bool]: A tuple containing:
            - The resolved callable provider.
            - A boolean indicating if the provider was lazily loaded (nested import).

    Raises:
        ImproperlyConfigured: If the provider is not a string.
        ImproperlyConfigured: If the resolved object is not callable.
    """
    if not isinstance(provider, str):
        raise ImproperlyConfigured(
            "The `provider` must be a string with the format '<module>.<callable>'."
        )

    is_nested: bool = False
    provider_callable: Callable

    try:
        provider_callable = load(provider)
    except ModuleNotFoundError:
        provider_callable = _get_provider_callable(provider)
        is_nested = True

    if not callable(provider_callable):
        raise ImproperlyConfigured(
            f"The `provider` must be callable, but `{provider}` is of type `{type(provider_callable)}`."
        )

    return cast(Callable, provider_callable), is_nested



from typing import Any, Optional

from lilya.exceptions import HTTPException
from lilya.requests import Request
from lilya.status import HTTP_403_FORBIDDEN
from typing_extensions import Annotated, Doc

from esmerald.openapi.models import OpenIdConnect as OpenIdConnectModel
from esmerald.security.base import SecurityBase as SecurityBase


class OpenIdConnect(SecurityBase):
    def __init__(
        self,
        *,
        openIdConnectUrl: Annotated[
            str,
            Doc(
                """
                The OpenID Connect URL.
                """
            ),
        ],
        scheme_name: Annotated[
            Optional[str],
            Doc(
                """
                The name of the security scheme.
                """
            ),
        ] = None,
        description: Annotated[
            Optional[str],
            Doc(
                """
                A description of the security scheme.
                """
            ),
        ] = None,
        auto_error: Annotated[
            bool,
            Doc(
                """
                Determines the behavior when the HTTP Authorization header is missing.

                If set to `True` (default), the request will be automatically canceled and an error will be sent to the client if the header is not provided.

                If set to `False`, the dependency result will be `None` when the header is not available, allowing for optional authentication or multiple authentication methods (e.g., OpenID Connect or a cookie).
                """
            ),
        ] = True,
    ):
        model = OpenIdConnectModel(
            openIdConnectUrl=openIdConnectUrl, description=description, scheme=scheme_name
        )
        model_dump = model.model_dump()
        super().__init__(**model_dump)
        self.scheme_name = scheme_name or self.__class__.__name__
        self.__auto_error__ = auto_error

    async def __call__(self, request: Request) -> Any:
        authorization = request.headers.get("Authorization")

        if authorization:
            return authorization

        if self.__auto_error__:
            raise HTTPException(status_code=HTTP_403_FORBIDDEN, detail="Not authenticated")

        return None



from typing import Union, cast

from lilya.exceptions import HTTPException
from lilya.status import HTTP_403_FORBIDDEN
from pydantic import BaseModel
from typing_extensions import Annotated, Doc

from esmerald.openapi.models import APIKey, APIKeyIn
from esmerald.requests import Request
from esmerald.security.base import SecurityBase


class APIKeyBase(SecurityBase):
    __model__: Union[BaseModel, None] = None


class APIKeyInQuery(APIKeyBase):
    """
    API key authentication using a query parameter.

    Defines the query parameter name for the API key and integrates it into the OpenAPI documentation.
    Extracts the key value from the query parameter and provides it as the dependency result.

    ## Usage

    Create an instance and use it as a dependency in `Inject()`.

    The dependency result will be a string containing the key value after using the `Injects()`.

    ## Example

    ```python
    from esmerald import Esmerald,  Gateway, get, Inject, Injects
    from esmerald.security.api_key import APIKeyInQuery

    query_scheme = APIKeyInQuery(name="api_key")

    @get("/items/", dependencies={"api_key": Inject(query_scheme)})
    async def read_items(api_key: str = Injects()) -> dict[str, str]:
        return {"api_key": api_key}
    ```
    """

    def __init__(
        self,
        *,
        name: Annotated[str, Doc("Name of the query parameter.")],
        scheme_name: Annotated[
            Union[str, None],
            Doc("Name of the security scheme, shown in OpenAPI documentation."),
        ] = None,
        description: Annotated[
            Union[str, None],
            Doc("Description of the security scheme, shown in OpenAPI documentation."),
        ] = None,
        auto_error: Annotated[
            bool,
            Doc(
                "If True, raises an error if the query parameter is missing. "
                "If False, returns None when the query parameter is missing."
            ),
        ] = True,
    ):
        model: APIKey = APIKey(
            **{"in": APIKeyIn.query.value},  # type: ignore[arg-type]
            name=name,
            description=description,
        )
        super().__init__(**model.model_dump())
        self.__model__ = model
        self.scheme_name = scheme_name or self.__class__.__name__
        self.__auto_error__ = auto_error

    async def __call__(self, request: Request) -> Union[str, None]:
        api_key = request.query_params.get(self.__model__.name)
        if api_key:
            return cast(str, api_key)
        if self.__auto_error__:
            raise HTTPException(status_code=HTTP_403_FORBIDDEN, detail="Not authenticated")
        return None


class APIKeyInHeader(APIKeyBase):
    """
    API key authentication using a header parameter.

    Defines the header parameter name for the API key and integrates it into the OpenAPI documentation.
    Extracts the key value from the header parameter and provides it as the dependency result.

    ## Usage

    Create an instance and use it as a dependency in `Inject()`.

    The dependency result will be a string containing the key value after using the `Injects()`.

    ## Example

    ```python
    from esmerald import Esmerald,  Gateway, get, Inject, Injects
    from esmerald.security.api_key import APIKeyInHeader

    header_scheme = APIKeyInHeader(name="x-key")

    @get("/items/", dependencies={"api_key": Inject(header_scheme)})
    async def read_items(api_key: str = Injects()) -> dict[str, str]:
        return {"api_key": api_key}
    ```
    """

    def __init__(
        self,
        *,
        name: Annotated[str, Doc("The name of the header parameter.")],
        scheme_name: Annotated[
            Union[str, None],
            Doc("The name of the security scheme to be shown in the OpenAPI documentation."),
        ] = None,
        description: Annotated[
            Union[str, None],
            Doc("A description of the security scheme to be shown in the OpenAPI documentation."),
        ] = None,
        auto_error: Annotated[
            bool,
            Doc(
                "If True, an error is raised if the header is missing. "
                "If False, None is returned when the header is missing."
            ),
        ] = True,
    ):
        model: APIKey = APIKey(
            **{"in": APIKeyIn.header.value},  # type: ignore[arg-type]
            name=name,
            description=description,
        )
        super().__init__(**model.model_dump())
        self.__model__ = model
        self.scheme_name = scheme_name or self.__class__.__name__
        self.__auto_error__ = auto_error

    async def __call__(self, request: Request) -> Union[str, None]:
        api_key = request.headers.get(self.__model__.name)
        if api_key:
            return cast(str, api_key)
        if self.__auto_error__:
            raise HTTPException(status_code=HTTP_403_FORBIDDEN, detail="Not authenticated")
        return None


class APIKeyInCookie(APIKeyBase):
    """
    API key authentication using a cookie parameter.

    Defines the cookie parameter name for the API key and integrates it into the OpenAPI documentation.
    Extracts the key value from the cookie parameter and provides it as the dependency result.

    ## Usage

    Create an instance and use it as a dependency in `Inject()`.

    The dependency result will be a string containing the key value after using the `Injects()`.

    ## Example

    ```python
    from esmerald import Esmerald,  Gateway, get, Inject, Injects
    from esmerald.security.api_key import APIKeyInCookie

    cookie_scheme = APIKeyInCookie(name="session")

    @get("/items/", dependencies={"api_key": Inject(cookie_scheme)})
    async def read_items(api_key: str = Injects()) -> dict[str, str]:
        return {"api_key": api_key}
    ```
    """

    def __init__(
        self,
        *,
        name: Annotated[str, Doc("The name of the cookie parameter.")],
        scheme_name: Annotated[
            Union[str, None],
            Doc("The name of the security scheme to be shown in the OpenAPI documentation."),
        ] = None,
        description: Annotated[
            Union[str, None],
            Doc("A description of the security scheme to be shown in the OpenAPI documentation."),
        ] = None,
        auto_error: Annotated[
            bool,
            Doc(
                "If True, an error is raised if the cookie is missing. "
                "If False, None is returned when the cookie is missing."
            ),
        ] = True,
    ):
        model: APIKey = APIKey(
            **{"in": APIKeyIn.cookie.value},  # type: ignore[arg-type]
            name=name,
            description=description,
        )
        super().__init__(**model.model_dump())
        self.__model__ = model
        self.scheme_name = scheme_name or self.__class__.__name__
        self.__auto_error__ = auto_error

    async def __call__(self, request: Request) -> Union[str, None]:
        api_key = request.cookies.get(self.__model__.name)
        if api_key:
            return api_key
        if self.__auto_error__:
            raise HTTPException(status_code=HTTP_403_FORBIDDEN, detail="Not authenticated")
        return None



from typing import TYPE_CHECKING, Any, Optional, Tuple, Type, Union

from esmerald.core.injector.provider import load_provider
from esmerald.core.transformers.signature import SignatureModel
from esmerald.parsers import ArbitraryHashableBaseModel
from esmerald.typing import Void
from esmerald.utils.helpers import is_async_callable

if TYPE_CHECKING:  # pragma: no cover
    from esmerald.typing import AnyCallable


class Factory:
    def __init__(self, provides: Union["AnyCallable", str], *args: Any, **kwargs: Any) -> None:
        """
        A dependency injection factory that supports both positional and keyword arguments.

        The provider can be passed as either:
        - A direct callable
        - A string reference to be dynamically imported

        Example Usage:
            dependencies = {
                "user": Factory(UserDAO, db_session=session, cache=cache)
            }
        """
        self.__args: Tuple[Any, ...] = args
        self.__kwargs: dict[str, Any] = kwargs
        self.is_nested: bool = False

        if isinstance(provides, str):
            self.provides, self.is_nested = load_provider(provides)
        else:
            self.provides = provides

    def set_args(self, *args: Any, **kwargs: Any) -> None:
        """Set or update arguments dynamically."""
        self.__args = args
        self.__kwargs = kwargs

    @property
    def cls(self) -> "AnyCallable":  # pragma: no cover
        """Return the provided class or function."""
        return self.provides

    async def __call__(self) -> Any:
        """
        Instantiates the provided class/function, handling both sync and async cases.

        Supports:
            - Nested imports (e.g., MyClass.func, MyClass.SubClass.func)
            - Both sync and async callables
            - Positional and keyword arguments

        Example:
            Factory(UserDAO, db_session=session)
        """
        if self.is_nested:
            self.provides = self.provides()

        if is_async_callable(self.provides):
            return await self.provides(*self.__args, **self.__kwargs)
        return self.provides(*self.__args, **self.__kwargs)


class Inject(ArbitraryHashableBaseModel):
    def __init__(self, dependency: "AnyCallable", use_cache: bool = False, **kwargs: Any):
        super().__init__(**kwargs)
        self.dependency = dependency
        self.signature_model: Optional["Type[SignatureModel]"] = None
        self.use_cache = use_cache
        self.value: Any = Void

    async def __call__(self, **kwargs: dict[str, Any]) -> Any:
        if self.use_cache and self.value is not Void:
            return self.value

        if is_async_callable(self.dependency):
            value = await self.dependency(**kwargs)
        else:
            value = self.dependency(**kwargs)

        if self.use_cache:
            self.value = value

        return value

    def __eq__(self, other: Any) -> bool:
        return other is self or (
            isinstance(other, self.__class__)
            and other.dependency == self.dependency
            and other.use_cache == self.use_cache
            and other.value == self.value
        )

    def __hash__(self) -> int:
        values: dict[str, Any] = {}
        for key, value in self.__dict__.items():
            values[key] = None
            if isinstance(value, (list, set)):
                values[key] = tuple(value)
            else:
                values[key] = value
        return hash((type(self),) + tuple(values))



from __future__ import annotations

import os
from typing import TYPE_CHECKING, Any, cast

from monkay import Monkay

if TYPE_CHECKING:
    from esmerald.conf.global_settings import EsmeraldSettings

ENVIRONMENT_VARIABLE = "ESMERALD_SETTINGS_MODULE"

monkay: Monkay[None, EsmeraldSettings] = Monkay(
    globals(),
    settings_path=lambda: os.environ.get(
        ENVIRONMENT_VARIABLE, "esmerald.conf.global_settings.EsmeraldSettings"
    ),
)


class SettingsForward:
    def __getattribute__(self, name: str) -> Any:
        return getattr(monkay.settings, name)

    def __setattr__(self, name: str, value: Any) -> None:
        return setattr(monkay.settings, name, value)


settings: EsmeraldSettings = cast("EsmeraldSettings", SettingsForward())


def reload_settings() -> None:
    """
    Reloads the global settings.
    """
    monkay.settings = os.environ.get(
        ENVIRONMENT_VARIABLE, "esmerald.conf.global_settings.EsmeraldSettings"
    )



from collections.abc import Sequence
from mimetypes import guess_type
from pathlib import PurePath
from typing import TYPE_CHECKING, Any, Optional, Union

from lilya.types import Receive, Scope, Send

from esmerald.responses.base import Response
from esmerald.utils.enums import MediaType

if TYPE_CHECKING:  # pragma: no cover
    from esmerald.background import BackgroundTask, BackgroundTasks
    from esmerald.core.protocols.template import TemplateEngineProtocol
    from esmerald.encoders import Encoder
    from esmerald.types import ResponseCookies


class TemplateResponse(Response):
    def __init__(
        self,
        template_name: str,
        template_engine: "TemplateEngineProtocol",
        status_code: int = 200,
        context: Optional[dict[str, Any]] = None,
        background: Optional[Union["BackgroundTask", "BackgroundTasks"]] = None,
        headers: Optional[dict[str, Any]] = None,
        cookies: Optional["ResponseCookies"] = None,
        media_type: Union[MediaType, str] = MediaType.JSON,
        encoders: Union[Sequence["Encoder"], None] = None,
        passthrough_body_types: Union[tuple[type, ...], None] = None,
    ):
        if media_type == MediaType.JSON:  # we assume this is the default
            suffixes = PurePath(template_name).suffixes
            for suffix in suffixes:
                _type = guess_type(f"name{suffix}")[0]
                if _type:
                    media_type = _type
                    break
            else:
                media_type = MediaType.TEXT  # pragma: no cover

        self.template = template_engine.get_template(template_name)
        self.context = context or {}
        content = getattr(self.template, template_engine.get_template_render_function())(**context)
        super().__init__(
            content=content,
            status_code=status_code,
            headers=headers,
            media_type=media_type,
            background=background,
            cookies=cookies,
            encoders=encoders,
            passthrough_body_types=passthrough_body_types,
        )

    def make_response(self, content: Any) -> bytes:
        # ensure template string is not mangled
        if isinstance(content, str):
            content = content.encode(self.charset)
        return super().make_response(content)

    async def __call__(
        self, scope: Scope, receive: Receive, send: Send
    ) -> None:  # pragma: no cover
        request = self.context.get("request", {})
        extensions = request.get("extensions", {})
        if "http.response.template" in extensions:
            await send(
                {
                    "type": "http.response.template",
                    "template": self.template,
                    "context": self.context,
                }
            )
        await super().__call__(scope, receive, send)



import os
import sys
from pathlib import Path
from typing import TYPE_CHECKING, Annotated

import click
from rich.tree import Tree
from sayer import Argument, Option, command, error

from esmerald.core.directives.env import DirectiveEnv
from esmerald.core.directives.exceptions import DirectiveError
from esmerald.core.terminal import Terminal
from esmerald.core.terminal.utils import get_log_config, get_ui_toolkit

if TYPE_CHECKING:
    pass

terminal = Terminal()


def get_app_tree(module_paths: list[Path], discovery_file: str) -> Tree:
    """
    Generates a tree structure representing the application modules.
    """
    root = module_paths[0]
    name = f"{root.name}" if root.is_file() else f"📁 {root.name}"

    root_tree = Tree(name)

    if root.is_dir():
        init_path = root / "__init__.py"
        if init_path.is_file():
            root_tree.add("[dim]__init__.py[/dim]")
            root_tree.add(f"[dim]{discovery_file}[/dim]")

    tree = root_tree
    for sub_path in module_paths[1:]:
        sub_name = f"{sub_path.name}" if sub_path.is_file() else f"📁 {sub_path.name}"
        tree = tree.add(sub_name)
        if sub_path.is_dir():
            tree.add("[dim]__init__.py[/dim]")
            tree.add(f"[dim]{discovery_file}[/dim]")

    return root_tree


@command
def runserver(
    path: Annotated[
        str | None,
        Argument(
            required=False,
            help="A path to a Python file or package directory with ([blue]__init__.py[/blue] files) containing a [bold]Lilya[/bold] app. If not provided, Esmerald will try to discover.",
        ),
    ],
    port: Annotated[
        int, Option(8000, "-p", help="Port to run the development server.", show_default=True)
    ],
    reload: Annotated[
        bool, Option(False, "-r", help="Reload server on file changes.", show_default=True)
    ],
    host: Annotated[
        str, Option(default="localhost", help="Host to run the server on.", show_default=True)
    ],
    debug: Annotated[
        bool, Option(default=True, help="Run the server in debug mode.", show_default=True)
    ],
    log_level: Annotated[
        str, Option(default="debug", help="Log level for the server.", show_default=True)
    ],
    lifespan: Annotated[
        str, Option(default="on", help="Enable lifespan events.", show_default=True)
    ],
    settings: Annotated[
        str | None,
        Option(help="Any custom settings to be initialised.", required=False, show_default=False),
    ],
    proxy_headers: Annotated[
        bool,
        Option(
            default=True,
            help="Enable/Disable X-Forwarded-Proto, X-Forwarded-For, X-Forwarded-Port to populate remote address info.",
            show_default=True,
        ),
    ],
    workers: Annotated[
        int | None,
        Option(
            default=None,
            help="Use multiple worker processes. Mutually exclusive with the --reload flag.",
            required=False,
        ),
    ],
) -> None:
    """Starts the Esmerald development server.

    The --app can be passed in the form of <module>.<submodule>:<app> or be set
    as environment variable ESMERALD_DEFAULT_APP.

    Alternatively, if none is passed, Esmerald will perform the application discovery.

    It is strongly advised not to run this command in any other environment but development.
    This was designed to facilitate the development environment and should not be used in production.

    How to run: `esmerald runserver`
    """
    ctx = click.get_current_context()
    env = ctx.ensure_object(DirectiveEnv)

    with get_ui_toolkit() as toolkit:
        toolkit.print(
            "[gray50]Identifying package structures based on directories with [green]__init__.py[/green] files[/gray50]"
        )

        if getattr(env, "app", None) is None:
            error(
                "You cannot specify a custom directive without specifying the --app or setting "
                "ESMERALD_DEFAULT_APP environment variable."
            )
            sys.exit(1)

        if settings is not None:
            os.environ.setdefault("ESMERALD_SETTINGS_MODULE", settings)

        try:
            import uvicorn
        except ImportError:
            raise DirectiveError(
                detail="Uvicorn needs to be installed to run Esmerald `runserver`."
            ) from None

        server_environment: str = ""
        if os.environ.get("ESMERALD_SETTINGS_MODULE"):
            from esmerald.conf import settings as esmerald_settings

            server_environment = f"{esmerald_settings.environment} "

        if not server_environment:
            server_environment = "development"

        toolkit.print_title(f"[green]Starting {server_environment} server[/green]", tag="Esmerald")
        toolkit.print(f"Importing from [green]{env.command_path}[/green]", tag="Esmerald")
        toolkit.print(f"Importing module '{env.path}'", tag="Esmerald")
        toolkit.print_line()

        root_tree = get_app_tree(
            env.module_info.module_paths, discovery_file=env.module_info.discovery_file
        )

        toolkit.print(root_tree, tag="module")
        toolkit.print_line()

        toolkit.print(
            "[green]The [bold]Esmerald[/bold] object is imported using the following code:[/green]",
            tag="code",
        )
        toolkit.print(
            f"[underline]from [bold]{env.module_info.module_import[0]}[/bold] import [bold]{env.module_info.module_import[1]}[/bold]",
            tag=env.module_info.module_import[1],
        )

        # For the text access
        url = f"http://{host}:{port}"
        docs = f"http://{host}:{port}/docs/swagger"

        toolkit.print_line()
        toolkit.print(
            f"Server started at [link={url}]{url}[/]",
            tag="server",
        )
        toolkit.print(
            f"Visit the docs at [link={docs}]{docs}[/]",
            tag="docs",
        )

        app = env.app

        if os.environ.get("ESMERALD_SETTINGS_MODULE"):
            custom_message = f"'{os.environ['ESMERALD_SETTINGS_MODULE']}'"
            toolkit.print(
                f"Using custom settings module: [bold][green]{custom_message}[/green][/bold]",
                tag="settings",
            )
        else:
            from esmerald.conf import settings as esmerald_settings

            toolkit.print(
                f"Using default settings module: [bold][green]{esmerald_settings.__class__.__module__}.Settings[/green][/bold]",
                tag="settings",
            )

        toolkit.print(
            "[yellow][bold]Remember, [bold]runserver is for development purposes[/bold]. For production, use a proper ASGI server.[/bold][/yellow]",
            tag="note",
        )
        toolkit.print_line()

        if debug:
            app.debug = debug

        uvicorn.run(
            app=path or env.path,
            port=port,
            host=host,
            reload=reload,
            lifespan=lifespan,  # type: ignore
            proxy_headers=proxy_headers,
            workers=workers,
            log_level=log_level,
            log_config=get_log_config(),
        )



from typing import Any, Optional, Union

from pydantic import BaseModel, ConfigDict, Field

from .encoding import Encoding
from .example import Example
from .reference import Reference
from .schema import Schema


class MediaType(BaseModel):
    """Each Media Type Object provides schema and examples for the media type
    identified by its key."""

    media_type_schema: Optional[Union[Reference, Schema]] = Field(default=None, alias="schema")
    """
    The schema defining the content of the request, response, or parameter.
    """

    example: Optional[Any] = None
    """
    Example of the media type.

    The example object SHOULD be in the correct format as specified by the media type.

    The `example` field is mutually exclusive of the `examples` field.

    Furthermore, if referencing a `schema` which contains an example,
    the `example` value SHALL _override_ the example provided by the schema.
    """

    examples: Optional[dict[str, Union[Example, Reference]]] = None
    """
    Examples of the media type.

    Each example object SHOULD match the media type and specified schema if present.

    The `examples` field is mutually exclusive of the `example` field.

    Furthermore, if referencing a `schema` which contains an example,
    the `examples` value SHALL _override_ the example provided by the schema.
    """

    encoding: Optional[dict[str, Encoding]] = None
    """
    A map between a property name and its encoding information.
    The key, being the property name, MUST exist in the schema as a property.
    The encoding object SHALL only apply to `requestBody` objects
    when the media type is `multipart` or `application/x-www-form-urlencoded`.
    """

    model_config = ConfigDict(
        extra="ignore",
        populate_by_name=True,
        json_schema_extra={
            "examples": [
                {
                    "schema": {"$ref": "#/components/schemas/Pet"},
                    "examples": {
                        "cat": {
                            "summary": "An example of a cat",
                            "value": {
                                "name": "Fluffy",
                                "petType": "Cat",
                                "color": "White",
                                "gender": "male",
                                "breed": "Persian",
                            },
                        },
                        "dog": {
                            "summary": "An example of a dog with a cat's name",
                            "value": {
                                "name": "Puma",
                                "petType": "Dog",
                                "color": "Black",
                                "gender": "Female",
                                "breed": "Mixed",
                            },
                        },
                        "frog": {"$ref": "#/components/examples/frog-example"},
                    },
                }
            ]
        },
    )



from typing import Any, Callable, Optional, Sequence, Union

from pydantic.fields import AliasChoices, AliasPath

from esmerald import params
from esmerald.typing import Undefined
from esmerald.utils.enums import EncodingType

_PyUndefined: Any = Undefined


def Security(
    dependency: Optional[Callable[..., Any]] = None,
    *,
    scopes: Optional[Sequence[str]] = None,
    use_cache: bool = True,
) -> params.Security:
    """
    This function should be only called if Inject/Injects is not used in the dependencies.
    This is a simple wrapper of the classic Inject().
    """
    return params.Security(dependency=dependency, scopes=scopes, use_cache=use_cache)


def Requires(
    dependency: Optional[Callable[..., Any]] = None,
    *,
    use_cache: bool = True,
) -> Any:
    """
    This function should be only called if Inject/Injects is not used in the dependencies.
    This is a simple wrapper of the classic Depends().
    """
    return params.Requires(dependency=dependency, use_cache=use_cache)


def Form(
    default: Any = _PyUndefined,
    *,
    annotation: Optional[Any] = None,
    default_factory: Optional[Callable[..., Any]] = _PyUndefined,
    allow_none: Optional[bool] = True,
    media_type: Union[str, EncodingType] = EncodingType.URL_ENCODED,
    content_encoding: Optional[str] = None,
    alias: Optional[str] = None,
    alias_priority: Optional[int] = _PyUndefined,
    title: Optional[str] = None,
    description: Optional[str] = None,
    gt: Optional[float] = None,
    ge: Optional[float] = None,
    lt: Optional[float] = None,
    le: Optional[float] = None,
    min_length: Optional[int] = None,
    max_length: Optional[int] = None,
    pattern: Optional[str] = None,
    examples: Optional[list[Any]] = None,
    validation_alias: Optional[Union[str, AliasPath, AliasChoices]] = None,
    discriminator: Optional[str] = None,
    max_digits: Optional[int] = _PyUndefined,
    strict: Optional[bool] = _PyUndefined,
    frozen: Optional[bool] = None,
    validate_default: bool = True,
    init_var: bool = True,
    kw_only: bool = True,
    include_in_schema: bool = True,
    json_schema_extra: Optional[dict[str, Any]] = None,
) -> Any:
    return params.Form(
        default=default,
        annotation=annotation,
        default_factory=default_factory,
        allow_none=allow_none,
        media_type=media_type,
        content_encoding=content_encoding,
        alias=alias,
        alias_priority=alias_priority,
        title=title,
        description=description,
        gt=gt,
        ge=ge,
        lt=lt,
        le=le,
        min_length=min_length,
        max_length=max_length,
        pattern=pattern,
        examples=examples,
        validation_alias=validation_alias,
        discriminator=discriminator,
        max_digits=max_digits,
        strict=strict,
        frozen=frozen,
        validate_default=validate_default,
        init_var=init_var,
        kw_only=kw_only,
        include_in_schema=include_in_schema,
        json_schema_extra=json_schema_extra,
    )



from __future__ import annotations

from typing import Annotated

from sayer import Argument, Option, command, error, success

from esmerald.core.directives.exceptions import DirectiveError
from esmerald.core.directives.operations._constants import SECRET_KEY_INSECURE_PREFIX
from esmerald.core.directives.templates import TemplateDirective
from esmerald.utils.crypto import get_random_secret_key


@command(name="createapp")  # type: ignore
def create_app(
    name: Annotated[str, Argument(help="The name of the app.")],
    verbosity: Annotated[int, Option(1, "-v", help="Displays the files generated")],
    with_basic_controller: Annotated[
        bool, Option(False, help="Should generate a basic controller")
    ],
    context: Annotated[
        bool,
        Option(
            False,
            help="Should generate an application with a controller, service, repository, dto.",
        ),
    ],
) -> None:
    """Creates the scaffold of an application

    How to run: `esmerald createapp <NAME>`

    Example: `esmerald createapp myapp`
    """
    options = {
        "secret_key": SECRET_KEY_INSECURE_PREFIX + get_random_secret_key(),
        "verbosity": verbosity,
        "with_basic_controller": with_basic_controller,
        "simple": False,
        "app_context": context,
    }
    directive = TemplateDirective()

    try:
        directive.handle("app", name=name, **options)
        success(f" App {name} generated successfully!")
    except DirectiveError as e:
        error(str(e))



import inspect
import os
import sys
import typing
from functools import wraps
from typing import Callable, TypeVar

import click
from sayer import Argument, Option, Sayer, error
from sayer.core.groups.sayer import SayerGroup

from esmerald import __version__
from esmerald.core.directives.constants import (
    EXCLUDED_DIRECTIVES,
    HELP_PARAMETER,
    IGNORE_DIRECTIVES,
)
from esmerald.core.directives.env import DirectiveEnv
from esmerald.core.directives.operations._constants import ESMERALD_SETTINGS_MODULE  # noqa
from esmerald.core.directives.operations.createapp import create_app as create_app  # noqa
from esmerald.core.directives.operations.createdeployment import (
    create_deployment as create_deployment,  # noqa
)
from esmerald.core.directives.operations.createproject import (
    create_project as create_project,  # noqa
)
from esmerald.core.directives.operations.list import directives as directives  # noqa
from esmerald.core.directives.operations.run import run as run  # noqa
from esmerald.core.directives.operations.runserver import runserver as runserver  # noqa
from esmerald.core.directives.operations.shell import shell as shell  # noqa
from esmerald.core.directives.operations.show_urls import show_urls as show_urls  # noqa

T = TypeVar("T")


class DirectiveGroup(SayerGroup):
    """Custom directive group to handle with the context and directives commands"""

    def add_command(self, cmd: click.Command, name: str | None = None) -> None:
        if cmd.callback:
            cmd.callback = self.wrap_args(cmd.callback)
        return super().add_command(cmd, name)

    def wrap_args(self, func: Callable[..., T]) -> Callable[..., T]:
        original = inspect.unwrap(func)
        params = inspect.signature(original).parameters

        @wraps(func)
        def wrapped(ctx: click.Context, /, *args: typing.Any, **kwargs: typing.Any) -> T:
            scaffold = ctx.ensure_object(DirectiveEnv)
            if "env" in params:
                kwargs["env"] = scaffold
            return func(*args, **kwargs)

        # click.pass_context makes sure that 'ctx' is the first argument
        return click.pass_context(wrapped)

    def process_settings(self, ctx: click.Context) -> None:
        """
        Process the settings context" if any is passed.

        Exports any ESMERALD_SETTINGS_MODULE to the environment if --settings is passed and exists
        as one of the params of any subcommand.
        """
        args = [*ctx.protected_args, *ctx.args]
        cmd_name, cmd, args = self.resolve_command(ctx, args)
        sub_ctx = cmd.make_context(cmd_name, args, parent=ctx)

        settings = sub_ctx.params.get("settings", None)
        if settings:
            settings_module: str = os.environ.get(ESMERALD_SETTINGS_MODULE, settings)
            os.environ.setdefault(ESMERALD_SETTINGS_MODULE, settings_module)

    def invoke(self, ctx: click.Context) -> typing.Any:
        """
        Directives can be ignored depending of the functionality from what is being
        called.
        """
        path = ctx.params.get("app", None)

        # Process any settings
        self.process_settings(ctx)

        if HELP_PARAMETER not in sys.argv and not any(
            value in sys.argv for value in EXCLUDED_DIRECTIVES
        ):
            try:
                directive = DirectiveEnv()
                app_env = directive.load_from_env(path=path)
                ctx.obj = app_env
            except OSError as e:
                if not any(value in sys.argv for value in IGNORE_DIRECTIVES):
                    error(str(e))
                    sys.exit(1)
        return super().invoke(ctx)


help_text = """
Esmerald command line tool allowing to run Esmerald native directives or
project unique and specific directives by passing the `-n` parameter.

How to run Esmerald native: `esmerald createproject <NAME>`. Or any other Esmerald native command.

    Example: `esmerald createproject myapp`


How to run custom directives: `esmerald --app <APP-LOCATION> run -n <DIRECTIVE NAME> <ARGS>`.

    Example: `esmerald --app myapp:app run -n createsuperuser`

"""

esmerald_cli = Sayer(
    help=help_text,
    add_version_option=True,
    version=__version__,
    group_class=DirectiveGroup,
)


@esmerald_cli.callback(invoke_without_command=True)
def esmerald_callback(
    ctx: click.Context,
    name: typing.Annotated[str, Argument(help="The directive name.")],
    app: typing.Annotated[
        str,
        Option(
            required=False,
            help="Module path to the Esmerald application. In a module:path format.",
        ),
    ],
) -> None: ...


esmerald_cli.add_command(directives)
esmerald_cli.add_command(show_urls)
esmerald_cli.add_command(runserver)
esmerald_cli.add_command(run)
esmerald_cli.add_command(create_project)
esmerald_cli.add_command(create_app)
esmerald_cli.add_command(create_deployment)
esmerald_cli.add_command(shell)



from __future__ import annotations

import asyncio
import select
import sys
from collections.abc import Sequence
from contextvars import copy_context
from typing import Annotated, Any, Callable, Union

import click
from lilya._internal._events import AsyncLifespanContextManager  # noqa
from lilya.cli.directives.operations.shell.enums import ShellOption
from sayer import Option, command

from esmerald.core.directives.env import DirectiveEnv


@command
async def shell(
    env: DirectiveEnv,
    kernel: Annotated[
        str,
        Option(
            default="ipython",
            help="Which shell should start.",
            type=click.Choice(["ipython", "ptpython"]),
            show_default=True,
        ),
    ],
) -> None:
    """
    Starts an interactive ipython shell with all the models
    and important python libraries.
    """
    if (
        sys.platform != "win32"
        and not sys.stdin.isatty()
        and select.select([sys.stdin], [], [], 0)[0]
    ):
        exec(sys.stdin.read(), globals())
        return

    on_startup = getattr(env.app, "on_startup", [])
    on_shutdown = getattr(env.app, "on_shutdown", [])
    lifespan = getattr(env.app, "lifespan", None)
    lifespan = handle_lifespan_events(
        on_startup=on_startup, on_shutdown=on_shutdown, lifespan=lifespan
    )
    await run_shell(env.app, lifespan, kernel)
    return None


async def run_shell(app: Any, lifespan: Any, kernel: str) -> None:
    """Executes the database shell connection"""

    async with lifespan(app):
        if kernel == ShellOption.IPYTHON:
            from esmerald.core.directives.operations.shell.ipython import get_ipython

            ipython_shell = get_ipython(app=app)
            await asyncio.to_thread(copy_context().run, ipython_shell)
        else:
            from esmerald.core.directives.operations.shell.ptpython import get_ptpython

            ptpython = get_ptpython(app=app)
            await asyncio.to_thread(copy_context().run, ptpython)


def handle_lifespan_events(
    on_startup: Union[Sequence[Callable], None] = None,
    on_shutdown: Union[Sequence[Callable], None] = None,
    lifespan: Union[Any, None] = None,
) -> Any:
    if lifespan:
        return lifespan
    return AsyncLifespanContextManager(on_startup=on_startup, on_shutdown=on_shutdown)



import functools
import math
import warnings
from typing import Any, Callable, Optional, Sequence, Union

import bcrypt
from monkay import load

from esmerald.conf import settings
from esmerald.exceptions import ImproperlyConfigured
from esmerald.utils.crypto import get_random_string as _get_random_string

from .constants import (
    RANDOM_STRING_CHARS,
    UNUSABLE_PASSWORD_PREFIX,
    UNUSABLE_PASSWORD_SUFFIX_LENGTH,
)


def is_password_usable(encoded: Optional[str]) -> bool:
    """
    Return True if this password wasn't generated by
    User.set_unusable_password(), i.e. make_password(None).
    """
    return encoded is None or not encoded.startswith(UNUSABLE_PASSWORD_PREFIX)


async def check_password(
    password: str,
    encoded: str,
    setter: Callable[..., Any] = None,
    preferred: str = "default",
) -> bool:
    """
    Return a boolean of whether the raw password matches the three
    part encoded.

    If setter is specified, it'll be called when you need to
    regenerate the password.
    """
    if password is None or not is_password_usable(encoded):
        return False

    preferred_hasher: BasePasswordHasher = get_hasher(preferred)
    try:
        hasher_handler = identify_hasher(encoded)
    except ValueError:
        # encoded is gibberish or uses a hasher that's no longer installed.
        return False

    hasher_changed = hasher_handler.algorithm != preferred_hasher.algorithm
    must_update: bool = hasher_changed or preferred_hasher.must_update(encoded)
    is_correct: bool = hasher_handler.verify(password, encoded)

    if setter and is_correct and must_update:
        await setter(password)
    return is_correct


def make_password(password: Optional[str], hasher: str = "default") -> str:
    """
    Turn a plain-text password into a hash for database storage

    Same as encode() but generate a new random salt. If password is None then
    return a concatenation of UNUSABLE_PASSWORD_PREFIX and a random string,
    which disallows logins. Additional random string reduces chances of gaining
    access to staff or superuser accounts. See ticket #20079 for more info.
    """
    if password is None:
        return UNUSABLE_PASSWORD_PREFIX + get_random_string(UNUSABLE_PASSWORD_SUFFIX_LENGTH)
    if not isinstance(password, (bytes, str)):
        raise TypeError(
            "Password must be a string or bytes, got %s." % type(password).__qualname__
        )

    hasher_handler: BasePasswordHasher = get_hasher(hasher)

    return hasher_handler.get_hashed_password(password)


@functools.lru_cache
def get_hashers() -> Sequence["BasePasswordHasher"]:
    hashers: Sequence["BasePasswordHasher"] = []

    password_hashers = getattr(settings, "password_hashers", None)
    if not password_hashers:
        warnings.warn("`password_hashers` missing from settings.", stacklevel=2)
        return hashers

    for hasher_path in settings.password_hashers:
        hasher_cls = load(hasher_path)
        hasher = hasher_cls()
        if not hasher.algorithm:
            raise ImproperlyConfigured(
                "hasher doesn't specify an algorithm name: %s" % hasher_path
            )
        hashers.append(hasher)
    return hashers


@functools.lru_cache
def get_hashers_by_algorithm() -> dict[str, "BasePasswordHasher"]:
    return {hasher.algorithm_name: hasher for hasher in get_hashers()}


def must_update_salt(salt: str, expected_entropy: int) -> bool:
    # Each character in the salt provides log_2(len(alphabet)) bits of entropy.
    return len(salt) * math.log2(len(RANDOM_STRING_CHARS)) < expected_entropy


def get_random_string(length: int, allowed_chars: str = RANDOM_STRING_CHARS) -> str:
    """
    Return a securely generated random string.

    The bit length of the returned value can be calculated with the formula:
        log_2(len(allowed_chars)^length)
    """
    return _get_random_string(length=length, allowed_chars=allowed_chars)


def get_hasher(algorithm: str = "default") -> "BasePasswordHasher":
    """
    Return an instance of a loaded password hasher.

    If algorithm is 'default', return the default hasher. Lazily import hashers
    specified in the project's settings config if needed.
    """
    if algorithm == "default":
        return get_hashers()[0]

    else:
        hashers = get_hashers_by_algorithm()
        try:
            return hashers[algorithm]
        except KeyError as e:
            raise ValueError(
                "Unknown password hashing algorithm '%s'. "
                "Did you specify it in your settings?" % algorithm
            ) from e


def identify_hasher(encoded: str) -> "BasePasswordHasher":
    """
    Return an instance of a loaded password hasher.

    Identify hasher algorithm by examining the encoded hash and call
    get_hasher() to return the correct hasher. Raise ValueError if
    algorithm cannot be identified, or if hasher is not loaded.
    """
    if encoded.startswith("$2b$") or encoded.startswith("$2a$") or encoded.startswith("$2y$"):
        # Identify bcrypt hash format
        algorithm = "bcrypt"
    else:
        # Extract algorithm identifier from the encoded string (split by '$' and get the first part)
        algorithm = encoded.split("$", 1)[0]

    return get_hasher(algorithm)


class BasePasswordHasher:
    hasher: Any
    algorithm: str
    algorithm_name: str
    salt_entropy: int = 128

    def __init__(self, **kwargs: Any) -> None:
        if self.algorithm is None or self.algorithm_name is None:
            raise NotImplementedError(
                "subclasses of BasePasswordHasher must provide an algorithm and algorithm_name."
            )

    def get_hashed_password(self, password: str) -> Union[str, Any]:
        # Use bcrypt to hash the password
        salt = bcrypt.gensalt().decode("utf-8")
        hashed_password = bcrypt.hashpw(password.encode("utf-8"), salt.encode("utf-8"))
        return hashed_password.decode("utf-8")

    def decode(self, encoded: str) -> dict[str, Any]:
        """
        Since bcrypt includes the salt in the hash, we don't need to manually decode.
        """
        return {"algorithm": self.algorithm_name, "hash": encoded}

    def must_update(self, encoded: str) -> bool:
        # bcrypt hashes don't require manual salt updates
        return False


class BcryptPasswordHasher(BasePasswordHasher):
    """
    Handles bcrypt password hashing in a similar way to PBKDF2PasswordHasher.
    """

    algorithm = "bcrypt"
    algorithm_name = "bcrypt"
    salt_entropy: int = 128
    rounds: int = 10  # Default cost factor for bcrypt

    def verify(self, password: str, encoded: str) -> bool:
        """
        Verify the password against the encoded hash.
        """
        # Use bcrypt's checkpw to verify password against full encoded hash
        return bcrypt.checkpw(password.encode("utf-8"), encoded.encode("utf-8"))

    def get_hashed_password(self, password: str) -> Union[str, Any]:
        """
        Hashes the password using bcrypt, returning a hash in the passlib-compatible format.
        """
        # Generate bcrypt salt
        salt = bcrypt.gensalt(rounds=self.rounds).decode("utf-8")
        # Hash the password with the generated salt
        hashed_password = bcrypt.hashpw(password.encode("utf-8"), salt.encode("utf-8"))
        return hashed_password.decode("utf-8")  # Return full bcrypt hash (salt + hash)

    def decode(self, encoded: str) -> dict[str, Any]:
        """
        Decode the bcrypt hash, splitting the salt and hash components.
        """
        salt, _hash = encoded[:29], encoded[29:]
        return {
            "algorithm": self.algorithm_name,
            "hash": _hash,
            "salt": salt,
            "rounds": self.rounds,  # Include the rounds cost factor
        }

    def must_update(self, encoded: str) -> bool:
        """
        Check if the bcrypt hash requires an update, such as updating salt entropy or rounds.
        """
        decoded = self.decode(encoded)
        update_salt = must_update_salt(decoded["salt"], self.salt_entropy)

        return update_salt or decoded["rounds"] != self.rounds



from functools import partial
from typing import Any

import orjson
from lilya.responses import RESPONSE_TRANSFORM_KWARGS

from .json import BaseJSONResponse
from .mixins import ORJSONTransformMixin

try:
    import ujson
except ImportError:  # pragma: no cover
    ujson = None


class ORJSONResponse(ORJSONTransformMixin, BaseJSONResponse):
    """
    An alternative to `JSONResponse` and performance wise, faster.

    In the same way the JSONResponse is used, so is the `ORJSONResponse`.
    """

    def make_response(self, content: Any) -> bytes:
        new_params = RESPONSE_TRANSFORM_KWARGS.get()
        if new_params:
            new_params = new_params.copy()
        else:
            new_params = {}
        new_params.setdefault(
            "json_encode_fn",
            partial(
                orjson.dumps,
                option=orjson.OPT_SERIALIZE_NUMPY | orjson.OPT_OMIT_MICROSECONDS,
            ),
        )
        with self.with_transform_kwargs(new_params):
            return super().make_response(content)


class UJSONResponse(BaseJSONResponse):
    """
    An alternative to `JSONResponse` and performance wise, faster.

    In the same way the JSONResponse is used, so is the `UJSONResponse`.
    """

    def make_response(self, content: Any) -> bytes:
        assert ujson is not None, "You must install the encoders or ujson to use UJSONResponse"
        # UJSON is actually in maintainance mode, recommends switch to ORJSON
        # https://github.com/ultrajson/ultrajson
        return ujson.dumps(content, ensure_ascii=False).encode("utf-8")


__all__ = ["ORJSONResponse", "UJSONResponse"]



from __future__ import annotations

from typing import Any, cast

from lilya.middleware.base import DefineMiddleware


def wrap_middleware(middleware: DefineMiddleware | str | Any, **kwargs: Any) -> DefineMiddleware:
    """
    Wraps a middleware input into a `DefineMiddleware` instance, ensuring consistent middleware usage.

    This utility function is designed to accept a middleware in various formats—either as an already
    instantiated `DefineMiddleware`, a string path to a middleware class, or a raw class/function reference—
    and normalize it into a `DefineMiddleware` object. This allows for flexible middleware registration
    while maintaining a consistent internal representation.

    ### Parameters
    ----------
    middleware : DefineMiddleware | str | Any
        The middleware to wrap. This can be:
        - An instance of `DefineMiddleware`, which will be returned directly.
        - A string representing the import path of a middleware class/function.
        - A class or function that implements the ASGI middleware interface.

    **kwargs : Any
        Additional keyword arguments to be passed to the `DefineMiddleware` constructor if the
        middleware is not already a `DefineMiddleware` instance.

    ### Returns
    -------
    DefineMiddleware
        A properly instantiated `DefineMiddleware` object representing the middleware.
        If the input is already an instance, it is returned as-is. Otherwise, a new instance is created
        using the given middleware and any additional keyword arguments.

    ### Example
    -------
    >>> from lilya.middleware.base import DefineMiddleware
    >>> class MyMiddleware:
    ...     def __init__(self, app):
    ...         self.app = app
    ...     async def __call__(self, scope, receive, send):
    ...         await self.app(scope, receive, send)

    >>> wrap_middleware(MyMiddleware, some_arg="value")
    DefineMiddleware(MyMiddleware, some_arg="value")

    >>> wrap_middleware("myproject.middleware.MyMiddleware", some_arg="value")
    DefineMiddleware("myproject.middleware.MyMiddleware", some_arg="value")

    >>> already_wrapped = DefineMiddleware(MyMiddleware)
    >>> wrap_middleware(already_wrapped) is already_wrapped
    True
    """
    if isinstance(middleware, DefineMiddleware):
        return middleware
    return DefineMiddleware(cast(Any, middleware), **kwargs)



from contextlib import suppress
from json import JSONDecodeError
from typing import TYPE_CHECKING, Any, get_args, get_origin

from lilya.datastructures import DataUpload as LilyaUploadFile
from orjson import loads
from pydantic import BaseModel, ConfigDict
from pydantic.fields import FieldInfo

from esmerald.core.datastructures import UploadFile
from esmerald.utils.enums import EncodingType

if TYPE_CHECKING:  # pragma: no cover
    from lilya.datastructures import FormData


class HashableBaseModel(BaseModel):  # pragma: no cover
    """
    Pydantic BaseModel by default doesn't handle with hashable types the same way
    a python object would and therefore there are types that are mutable (list, set)
    not hashable and those need to be handled properly.

    HashableBaseModel handles those corner cases.
    """

    def __hash__(self) -> int:
        values: dict[str, Any] = {}
        for key, value in self.__dict__.items():
            values[key] = None
            if isinstance(value, (list, set)):
                values[key] = tuple(value)
            else:
                values[key] = value
        return hash((type(self),) + tuple(values))


class ArbitraryHashableBaseModel(HashableBaseModel):
    """
    Same as HashableBaseModel but allowing arbitrary values
    """

    model_config = ConfigDict(extra="allow", arbitrary_types_allowed=True)


class ArbitraryBaseModel(BaseModel):
    """
    ArbitraryBaseModel that allows arbitrary_types_allowed to be passed.
    """

    model_config = ConfigDict(arbitrary_types_allowed=True)


class ArbitraryExtraBaseModel(BaseModel):
    model_config = ConfigDict(extra="allow", arbitrary_types_allowed=True)


def flatten(values: list[Any]) -> list[Any]:
    """
    Flattens a list
    """
    flattened = []

    for value in values:
        if isinstance(value, list):
            flattened.extend(flatten(value))
        else:
            flattened.append(value)
    return flattened


def parse_json_value(value: Any) -> Any:
    """
    Attempts to parse a JSON string into a Python object.

    If parsing fails, returns the original value.

    Args:
        value (Any): The value to be parsed.

    Returns:
        Any: Parsed JSON object or the original value.
    """
    if not isinstance(value, str):
        return value

    with suppress(JSONDecodeError):
        return loads(value)

    return value  # type: ignore


def merge_values(values_dict: dict[str, Any], key: str, value: Any) -> None:
    """
    Merges values into a dictionary, ensuring that multiple values for the same key
    are stored in a list.

    Args:
        values_dict (dict[str, Any]): The dictionary storing parsed form data.
        key (str): The key for the form field.
        value (Any): The value to be added.
    """
    if key in values_dict:
        existing_value = values_dict[key]
        if isinstance(existing_value, list):
            existing_value.append(value)
        else:
            values_dict[key] = [existing_value, value]
    else:
        values_dict[key] = value


def process_form_data(form_data: "FormData") -> dict[str, Any]:
    """
    Processes form data, converting JSON-encoded strings and organizing values into a dictionary.

    Args:
        form_data (FormData): The form data to be processed.

    Returns:
        dict[str, Any]: Parsed and structured form data.
    """
    values_dict: dict[str, Any] = {}

    for key, value in form_data.multi_items():
        parsed_value = value if isinstance(value, LilyaUploadFile) else parse_json_value(value)
        merge_values(values_dict, key, parsed_value)

    return values_dict


def handle_multipart_encoding(field: "FieldInfo", values_dict: dict[str, Any]) -> Any:
    """
    Handles parsing of multipart form data based on field annotations.

    Args:
        field (FieldInfo): Field metadata including expected data types.
        values_dict (dict[str, Any]): Parsed form data.

    Returns:
        Any: The processed value, formatted correctly according to field type.
    """
    if not values_dict:
        return None

    # If the field expects a list, flatten the values
    if get_origin(field.annotation) is list:
        return flatten(list(values_dict.values()))

    # Handle direct UploadFile types
    if field.annotation in (LilyaUploadFile, UploadFile):
        return list(values_dict.values())[0]

    # Check if UploadFile is inside a Union (e.g., Optional[UploadFile])
    for arg in get_args(field.annotation):
        if issubclass(arg, (LilyaUploadFile, UploadFile)):
            return list(values_dict.values())[0]

    return values_dict


def parse_form_data(
    media_type: "EncodingType", form_data: "FormData", field: "FieldInfo"
) -> Any:  # pragma: no cover
    """
    Parses form data by converting JSON-encoded fields, merging multiple values,
    and handling multipart encoding appropriately.

    Args:
        media_type (EncodingType): The encoding type of the form data.
        form_data (FormData): The form data containing key-value pairs.
        field (FieldInfo): Metadata defining the expected data structure.

    Returns:
        Any: Parsed form data in an appropriate format, based on the field type.
    """
    values_dict = process_form_data(form_data)

    if media_type == EncodingType.MULTI_PART:
        return handle_multipart_encoding(field, values_dict)

    return values_dict if values_dict else None



from functools import partial
from typing import Any

import orjson
from lilya.responses import RESPONSE_TRANSFORM_KWARGS


class ORJSONTransformMixin:
    @classmethod
    def transform(cls, value: Any) -> Any:
        """
        The transformation of the data being returned (simplify operation).

        Supports all the default encoders from Lilya and custom from Esmerald.
        """
        transform_kwargs = RESPONSE_TRANSFORM_KWARGS.get()
        if transform_kwargs is None:
            transform_kwargs = {}
        else:
            transform_kwargs.copy()
        transform_kwargs.setdefault(
            "json_encode_fn",
            partial(
                orjson.dumps, option=orjson.OPT_SERIALIZE_NUMPY | orjson.OPT_OMIT_MICROSECONDS
            ),
        )
        transform_kwargs.setdefault("post_transform_fn", orjson.loads)

        with cls.with_transform_kwargs(transform_kwargs):
            return super().transform(value)



from __future__ import annotations

from functools import partial
from typing import (
    TYPE_CHECKING,
    Any,
    Generic,
    NoReturn,
    Optional,
    Sequence,
    TypeVar,
    Union,
    cast,
)

import orjson
from lilya import status
from lilya.responses import (
    RESPONSE_TRANSFORM_KWARGS,
    Error as Error,  # noqa
    FileResponse as FileResponse,  # noqa
    HTMLResponse as HTMLResponse,  # noqa
    JSONResponse as JSONResponse,  # noqa
    Ok as Ok,  # noqa
    PlainText as PlainText,  # noqa
    RedirectResponse as RedirectResponse,  # noqa
    Response as LilyaResponse,  # noqa
    StreamingResponse as StreamingResponse,  # noqa
)
from typing_extensions import Annotated, Doc

from esmerald.encoders import Encoder
from esmerald.exceptions import ImproperlyConfigured
from esmerald.utils.enums import MediaType

from .mixins import ORJSONTransformMixin

PlainTextResponse = PlainText

if TYPE_CHECKING:  # pragma: no cover
    from esmerald.background import BackgroundTask, BackgroundTasks
    from esmerald.types import ResponseCookies

T = TypeVar("T")


class Response(ORJSONTransformMixin, LilyaResponse, Generic[T]):
    """
    Default `Response` object from Esmerald where it can be as the
    return annotation of a [handler](https://esmerald.dev/routing/handlers/).

    Esmerakd automatically will understand what time of response is going to be
    used and parse all the details automatically.

    **Example**

    ```python
    from pydantic import BaseModel

    from esmerald import Esmerald, Gateway, Response, get
    from esmerald.datastructures import Cookie


    @get(path="/me")
    async def home() -> Response:
        return Response(
            Item(id=1, sku="sku1238"),
            headers={"SKY-HEADER": "sku-xyz"},
            cookies=[Cookie(key="sku", value="a-value")],
        )


    Esmerald(routes=[Gateway(handler=home)])
    ```
    """

    def __init__(
        self,
        content: Annotated[
            Any,
            Doc(
                """
                Any content being sent to the response.
                """
            ),
        ],
        *,
        status_code: Annotated[
            int,
            Doc(
                """
                The response status code.
                """
            ),
        ] = status.HTTP_200_OK,
        media_type: Annotated[
            Optional[Union[MediaType, str]],
            Doc(
                """
                The media type used in the response.
                """
            ),
        ] = None,
        background: Annotated[
            Optional[Union[BackgroundTask, BackgroundTasks]],
            Doc(
                """
                Any instance of a [BackgroundTask or BackgroundTasks](https://esmerald.dev/background-tasks/).
                """
            ),
        ] = None,
        headers: Annotated[
            Optional[dict[str, Any]],
            Doc(
                """
                Any additional headers being passed to the response.
                """
            ),
        ] = None,
        cookies: Annotated[
            Optional[ResponseCookies],
            Doc(
                """
                A sequence of `esmerald.datastructures.Cookie` objects.

                Read more about the [Cookies](https://esmerald.dev/extras/cookie-fields/?h=responsecook#cookie-from-response-cookies).

                **Example**

                ```python
                from esmerald import Response
                from esmerald.datastructures import Cookie

                response_cookies=[
                    Cookie(
                        key="csrf",
                        value="CIwNZNlR4XbisJF39I8yWnWX9wX4WFoz",
                        max_age=3000,
                        httponly=True,
                    )
                ]

                Response(cookies=response_cookies)
                ```
                """
            ),
        ] = None,
        encoders: Annotated[
            Union[Sequence[Encoder], None],
            Doc(
                """
                A sequence of `esmerald.encoders.Encoder` type of objects to be used
                by the response object directly.

                **Example**

                ```python
                from esmerald import Response
                from esmerald.encoders import PydanticEncoder, MsgSpecEncoder

                response_cookies=[
                    encoders=[PydanticEncoder, MsgSpecEncoder]
                ]

                Response(cookies=response_cookies)
                ```
                """
            ),
        ] = None,
        passthrough_body_types: Annotated[
            Union[tuple[type, ...], None],
            Doc(
                """
            A tuple with types, which should be passed through directly to the ASGI server.
            By default only "bytes" are passed through but some ASGI servers can also handle
            memoryviews or strings.
            If set, it should include the "bytes" type.
            As an alternative you can subclass and set a ClassVar named passthrough_body_types.

            Be warned: this is highly ASGI server specific and may requires you to set the content-length header yourself.
            """
            ),
        ] = None,
    ) -> None:
        self.cookies = cookies or []
        super().__init__(
            content=content,
            status_code=status_code,
            headers=headers or {},
            media_type=media_type,
            background=cast("BackgroundTask", background),
            encoders=encoders,
            passthrough_body_types=passthrough_body_types,
        )

    def make_response(self, content: Any) -> bytes:
        if (
            content is None
            or content is NoReturn
            and (
                self.status_code < 100
                or self.status_code in {status.HTTP_204_NO_CONTENT, status.HTTP_304_NOT_MODIFIED}
            )
        ):
            return b""
        transform_kwargs = RESPONSE_TRANSFORM_KWARGS.get()
        if transform_kwargs:
            transform_kwargs = transform_kwargs.copy()
        elif isinstance(content, str) and self.media_type != MediaType.JSON:
            # treat strings special when not using json and disable mangling when no context is active.
            transform_kwargs = None
        else:
            transform_kwargs = {}
        if transform_kwargs is not None:
            transform_kwargs.setdefault(
                "json_encode_fn",
                partial(
                    orjson.dumps,
                    option=orjson.OPT_SERIALIZE_NUMPY | orjson.OPT_OMIT_MICROSECONDS,
                ),
            )
        try:
            # switch to a special mode for MediaType.JSON (default handlers)
            if self.media_type == MediaType.JSON:
                # keep it a serialized json object
                if transform_kwargs is not None:
                    transform_kwargs.setdefault("post_transform_fn", None)
            # otherwise use default logic of lilya striping '"'
            with self.with_transform_kwargs(transform_kwargs):
                # if content is bytes it won't be transformed and
                # if None or NoReturn, return b"", this differs from the dedicated JSONResponses.
                return super().make_response(content)
        except (AttributeError, ValueError, TypeError) as e:  # pragma: no cover
            raise ImproperlyConfigured("Unable to serialize response content") from e
